"""Testing cross-validation analysis."""

import warnings

import numpy as np
import numpy.testing as npt

import dipy.core.gradients as gt
import dipy.data as dpd
from dipy.io.image import load_nifti_data
import dipy.reconst.base as base
import dipy.reconst.cross_validation as xval
import dipy.reconst.csdeconv as csd
import dipy.reconst.dti as dti
from dipy.reconst.shm import descoteaux07_legacy_msg
import dipy.sims.voxel as sims
from dipy.testing.decorators import set_random_number_generator

# We'll set these globally:
fdata, fbval, fbvec = dpd.get_fnames(name="small_64D")


@set_random_number_generator()
def test_coeff_of_determination(rng):
    model = rng.standard_normal((10, 10, 10, 150))
    data = np.copy(model)
    # If the model predicts the data perfectly, the COD is all 100s:
    cod = xval.coeff_of_determination(data, model)
    npt.assert_array_equal(100, cod)


def test_dti_xval():
    data = load_nifti_data(fdata)
    gtab = gt.gradient_table(fbval, bvecs=fbvec)
    dm = dti.TensorModel(gtab, fit_method="LS")
    # The data has 102 directions, so will not divide neatly into 10 bits
    npt.assert_raises(ValueError, xval.kfold_xval, dm, data, 10)

    # In simulation with no noise, COD should be perfect:
    psphere = dpd.get_sphere(name="symmetric362")
    bvecs = np.concatenate(([[0, 0, 0]], psphere.vertices))
    bvals = np.zeros(len(bvecs)) + 1000
    bvals[0] = 0
    gtab = gt.gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0001], [0.0015, 0.0003, 0.0003]))
    mevecs = [
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
    ]
    S = sims.single_tensor(gtab, 100, evals=mevals[0], evecs=mevecs[0], snr=None)

    dm = dti.TensorModel(gtab, fit_method="LS")
    kf_xval = xval.kfold_xval(dm, S, 2)
    cod = xval.coeff_of_determination(S, kf_xval)
    npt.assert_array_almost_equal(cod, np.ones(kf_xval.shape[:-1]) * 100)

    # Test with 2D data for use of a mask
    S = np.array([[S, S], [S, S]])
    mask = np.ones(S.shape[:-1], dtype=bool)
    mask[1, 1] = 0
    kf_xval = xval.kfold_xval(dm, S, 2, mask=mask)
    cod2d = xval.coeff_of_determination(S, kf_xval)
    npt.assert_array_almost_equal(np.round(cod2d[0, 0]), cod)


@set_random_number_generator(12345)
def test_csd_xval(rng):
    # First, let's see that it works with some data:
    data = load_nifti_data(fdata)[1:3, 1:3, 1:3]  # Make it *small*
    gtab = gt.gradient_table(fbval, bvecs=fbvec)
    S0 = np.mean(data[..., gtab.b0s_mask])
    response = ([0.0015, 0.0003, 0.0001], S0)

    # In simulation, it should work rather well (high COD):
    psphere = dpd.get_sphere(name="symmetric362")
    bvecs = np.concatenate(([[0, 0, 0]], psphere.vertices))
    bvals = np.zeros(len(bvecs)) + 1000
    bvals[0] = 0
    gtab = gt.gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0001], [0.0015, 0.0003, 0.0003]))
    mevecs = [
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
    ]
    S0 = 100
    S = sims.single_tensor(gtab, S0, evals=mevals[0], evecs=mevecs[0], snr=None)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        sm = csd.ConstrainedSphericalDeconvModel(gtab, response)
    response = ([0.0015, 0.0003, 0.0001], S0)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        kf_xval = xval.kfold_xval(sm, S, 2, response, sh_order_max=2, rng=rng)
    # Because of the regularization, COD is not going to be perfect here:
    cod = xval.coeff_of_determination(S, kf_xval)
    # We'll just test for regressions:
    csd_cod = 97  # pre-computed by hand for this random seed

    # We're going to be really lenient here:
    npt.assert_array_almost_equal(np.round(cod), csd_cod)
    # Test for sD data with more than one voxel for use of a mask:
    S = np.array([[S, S], [S, S]])
    mask = np.ones(S.shape[:-1], dtype=bool)
    mask[1, 1] = 0
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        kf_xval = xval.kfold_xval(
            sm, S, 2, response, sh_order_max=2, mask=mask, rng=rng
        )

    cod = xval.coeff_of_determination(S, kf_xval)
    npt.assert_array_almost_equal(np.round(cod[0]), csd_cod)


def test_no_predict():
    # Test that if you try to do this with a model that doesn't have a `predict`
    # method, you get something reasonable.
    class NoPredictModel(base.ReconstModel):
        def __init__(self, gtab):
            base.ReconstModel.__init__(self, gtab)

        def fit(self, data, mask=None):
            return NoPredictFit(self, data, mask=mask)

    class NoPredictFit(base.ReconstFit):
        def __init__(self, model, data, mask=None):
            base.ReconstFit.__init__(self, model, data)

    gtab = gt.gradient_table(fbval, bvecs=fbvec)
    my_model = NoPredictModel(gtab)
    data = load_nifti_data(fdata)[1:3, 1:3, 1:3]  # Whatever

    npt.assert_raises(ValueError, xval.kfold_xval, my_model, data, 2)



"""Testing weights methods"""

import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
    assert_raises,
)

import dipy.core.gradients as grad
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
import dipy.reconst.dti as dti
from dipy.reconst.weights_method import (
    simple_cutoff,
    two_eyes_cutoff,
    weights_method_nlls_m_est,
    weights_method_wls_m_est,
)
from dipy.testing.decorators import set_random_number_generator

MIN_POSITIVE_SIGNAL = 0.0001


@set_random_number_generator()
def test_outlier_funcs(rng):
    """Test functions that define outliers."""

    # true signal
    b0 = 1000.0
    bval, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table(bval, bvecs=bvecs)
    B = bval[1]
    D_orig = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B
    design_matrix = dti.design_matrix(gtab)
    log_pred_sig = np.dot(design_matrix, D_orig)
    pred_sig = np.exp(log_pred_sig)

    # noisy signal
    scale = 1
    error = rng.normal(scale=scale, size=pred_sig.shape)
    Y = pred_sig + error
    Y[Y < MIN_POSITIVE_SIGNAL] = MIN_POSITIVE_SIGNAL
    Y[0] = Y[0] * 100  # make 1 signal into a super outlier

    # other args
    residuals = Y - pred_sig  # may differ from residuals due to signal clip
    log_residuals = np.log(Y) - log_pred_sig
    # make some fake leverages (they should sum to npar=7)
    leverages = np.ones_like(Y) * D_orig.shape[0] / Y.shape[0]
    C = scale  # since we added noise, just set C to scale

    for outlier_func in [simple_cutoff, two_eyes_cutoff]:
        # use extremely generous cutoff of 6; find only the super outlier
        outlier = outlier_func(
            residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff=6
        )
        assert_equal(outlier[0], True)
        assert_equal(outlier[1:], np.zeros_like(outlier[1:], dtype=bool))

        # make everything an outlier, by using a cut-off of zero
        outlier = outlier_func(
            residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff=0.0
        )
        assert_equal(outlier.sum(), Y.shape[0])


@set_random_number_generator()
def test_weights_funcs(rng):
    """Test functions that define weights."""

    # true signal
    b0 = 1000.0
    bval, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table(bval, bvecs=bvecs)
    B = bval[1]
    D_orig = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B
    design_matrix = dti.design_matrix(gtab)
    log_pred_sig = np.dot(design_matrix, D_orig)
    pred_sig = np.exp(log_pred_sig)

    # noisy signal
    scale = 1
    error = rng.normal(scale=scale, size=pred_sig.shape)
    Y = pred_sig + error
    Y[Y < MIN_POSITIVE_SIGNAL] = MIN_POSITIVE_SIGNAL
    Y[0] = Y[0] * 100  # make 1 signal into a super outlier

    # make some fake leverages (they should sum to npar=7)
    leverages = np.ones_like(Y) * D_orig.shape[0] / Y.shape[0]

    # set last robust to None ?
    # it's only currently used on the last iteration of WLS, because we need to
    # perform a clean OLS fit first
    # in general, weight functions for WLS will need to do this
    # makes sense that they generally have access to the defined 'last_robust'
    # from previous iterations, for the most generality

    # TODO: need to test about the number of iterations, to make sure errors are raised
    # TODO: robust return should be None expect on last for NLLS,
    #       and second to last and last on WLS

    NUM_ITER = 4
    for fit_type, weights_func in zip(
        ["wls", "nlls"], [weights_method_wls_m_est, weights_method_nlls_m_est]
    ):
        # invalid estimator choice
        assert_raises(
            ValueError,
            weights_func,
            Y,
            pred_sig,
            design_matrix,
            leverages,
            1,
            10,
            None,
            m_est="unknown_estimator",
        )

        # test error if <4 iters for wls, or <3 iters for nlls
        assert_raises(
            ValueError,
            weights_func,
            Y,
            pred_sig,
            design_matrix,
            leverages,
            1,
            3 if fit_type == "wls" else 2,
            None,
        )

        # test error if not enough data points
        assert_raises(
            ValueError,
            weights_func,
            Y[0:6],
            pred_sig[0:6],
            design_matrix[0:6],
            leverages[0:6],
            1,
            10,
            None,
        )

        for mest in ["gm", "cauchy"]:
            for outlier_func in [simple_cutoff, two_eyes_cutoff]:
                last_robust = None
                for idx in range(1, NUM_ITER + 1):
                    # neither last, nor second-to-last, iteration
                    weights, robust = weights_func(
                        Y,
                        pred_sig,
                        design_matrix,
                        leverages,
                        idx=idx,
                        total_idx=NUM_ITER,
                        last_robust=last_robust,
                        m_est=mest,
                        cutoff=3,
                        outlier_condition_func=outlier_func,
                    )
                    last_robust = robust

                    if idx >= NUM_ITER - 1 and fit_type == "wls":
                        # for second-to-last and last iter, robust not None
                        assert_equal(robust.shape, Y.shape)
                        # check that the super-outlier has 0 weight
                        assert_almost_equal(weights[0], 0)

                    if idx == NUM_ITER and fit_type == "nlls":
                        # for last iter, robust not None
                        assert_equal(robust.shape, Y.shape)
                        # check that the super-outlier has 0 weight
                        assert_almost_equal(weights[0], 0)

                    if idx < NUM_ITER - 1:  # iter okay for both wls & nlls
                        assert_equal(robust, None)
                        assert_equal(np.all(weights > 0), True)



"""
=========================
Getting started with DIPY
=========================

In diffusion MRI (dMRI) usually we use three types of files, a Nifti file with
the diffusion weighted data, and two text files one with b-values and
one with the b-vectors.

In DIPY_ we provide tools to load and process these files and we also provide
access to publicly available datasets for those who haven't acquired yet
their own datasets.

Let's start with some necessary imports.
"""

from os.path import expanduser, join

import matplotlib.pyplot as plt

from dipy.core.gradients import gradient_table
from dipy.data import fetch_sherbrooke_3shell
from dipy.io import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti

###############################################################################
# With the following commands we can download a dMRI dataset

fetch_sherbrooke_3shell()

###############################################################################
# By default these datasets will go in the ``.dipy`` folder inside your home
# directory. Here is how you can access them.

home = expanduser("~")

###############################################################################
# ``dname`` holds the directory name where the 3 files are in.

dname = join(home, ".dipy", "sherbrooke_3shell")

###############################################################################
# Here, we show the complete filenames of the 3 files

fdwi = join(dname, "HARDI193.nii.gz")

print(fdwi)

fbval = join(dname, "HARDI193.bval")

print(fbval)

fbvec = join(dname, "HARDI193.bvec")

print(fbvec)

###############################################################################
# Now, that we have their filenames we can start checking what these look like.
#
# Let's start first by loading the dMRI datasets. For this purpose, we
# use a python library called nibabel_ which enables us to read and write
# neuroimaging-specific file formats.

data, affine, img = load_nifti(fdwi, return_img=True)

###############################################################################
# ``data`` is a 4D array where the first 3 dimensions are the i, j, k voxel
# coordinates and the last dimension is the number of non-weighted (S0s) and
# diffusion-weighted volumes.
#
# We can very easily check the size of ``data`` in the following way:

print(data.shape)

###############################################################################
# We can also check the dimensions of each voxel in the following way:

print(img.header.get_zooms()[:3])

###############################################################################
# We can quickly visualize the results using matplotlib_. For example,
# let's show here the middle axial slices of volume 0 and volume 10.

axial_middle = data.shape[2] // 2
plt.figure("Showing the datasets")
plt.subplot(1, 2, 1).set_axis_off()
plt.imshow(data[:, :, axial_middle, 0].T, cmap="gray", origin="lower")
plt.subplot(1, 2, 2).set_axis_off()
plt.imshow(data[:, :, axial_middle, 10].T, cmap="gray", origin="lower")
plt.show()
plt.savefig("data.png", bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the middle axial slice without (left) and with (right) diffusion
# weighting.
#
#
#
# The next step is to load the b-values and b-vectors from the disk using
# the function ``read_bvals_bvecs``.

bvals, bvecs = read_bvals_bvecs(fbval, fbvec)

###############################################################################
# In DIPY, we use an object called ``GradientTable`` which holds all the
# acquisition specific parameters, e.g. b-values, b-vectors, timings and
# others. To create this object you can use the function ``gradient_table``.

gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# Finally, you can use ``gtab`` (the GradientTable object) to show some
# information about the acquisition parameters

print(gtab.info)

###############################################################################
# You can also see the b-values using:

print(gtab.bvals)

###############################################################################
# Or, for example the 10 first b-vectors using:

print(gtab.bvecs[:10, :])

###############################################################################
# You can get the number of gradients (including the number of b0 values)
# calling ``len`` on the ``GradientTable`` instance:

print(len(gtab))

###############################################################################
# ``gtab``  can be used to tell what part of the data is the S0 volumes
# (volumes which correspond to b-values of 0).

S0s = data[:, :, :, gtab.b0s_mask]

###############################################################################
# Here, we had only 1 S0 as we can verify by looking at the dimensions of S0s

print(S0s.shape)

###############################################################################
# Just, for fun let's save this in a new Nifti file.

save_nifti("HARDI193_S0.nii.gz", S0s, affine)

###############################################################################
# Now, that we learned how to load dMRI datasets we can start the analysis.
# See example :ref:`sphx_glr_examples_built_reconstruction_reconst_dti.py` to
# learn how to create FA maps.



from os.path import join as pjoin
import warnings

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package("fury", min_version="0.9.0")
if has_fury:
    from fury.colormap import colormap_lookup_table
    from fury.lib import (
        VTK_OBJECT,
        Actor,
        CellArray,
        Command,
        PolyData,
        PolyDataMapper,
        calldata_type,
        numpy_support,
    )
    from fury.shaders import (
        attribute_to_actor,
        compose_shader,
        import_fury_shader,
        shader_to_actor,
    )
    from fury.utils import apply_affine, numpy_to_vtk_colors, numpy_to_vtk_points
else:

    class Actor:
        pass

    def calldata_type(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def VTK_OBJECT(*args):
        pass


class PeakActor(Actor):
    """FURY actor for visualizing DWI peaks.

    Parameters
    ----------
    directions : ndarray
        Peak directions. The shape of the array should be (X, Y, Z, D, 3).
    indices : tuple
        Indices given in tuple(x_indices, y_indices, z_indices)
        format for mapping 2D ODF array to 3D voxel grid.
    values : ndarray, optional
        Peak values. The shape of the array should be (X, Y, Z, D).
    affine : array, optional
        4x4 transformation array from native coordinates to world coordinates.
    colors : None or string ('rgb_standard') or tuple (3D or 4D) or array/ndarray (N, 3 or 4) or (K, 3 or 4) or (N, ) or (K, )
        If None a standard orientation colormap is used for every line.
        If one tuple of color is used. Then all streamlines will have the same
        color.
        If an array (N, 3 or 4) is given, where N is equal to the number of
        points. Then every point is colored with a different RGB(A) color.
        If an array (K, 3 or 4) is given, where K is equal to the number of
        lines. Then every line is colored with a different RGB(A) color.
        If an array (N, ) is given, where N is the number of points then these
        are considered as the values to be used by the colormap.
        If an array (K,) is given, where K is the number of lines then these
        are considered as the values to be used by the colormap.
    lookup_colormap : vtkLookupTable, optional
        Add a default lookup table to the colormap. Look at
        :func:`fury.actor.colormap_lookup_table` for more information.
    linewidth : float, optional
        Line thickness.
    symmetric: bool, optional
        If True, peaks are drawn for both peaks_dirs and -peaks_dirs. Else,
        peaks are only drawn for directions given by peaks_dirs.

    """  # noqa: E501

    @warning_for_keywords()
    def __init__(
        self,
        directions,
        indices,
        *,
        values=None,
        affine=None,
        colors=None,
        lookup_colormap=None,
        linewidth=1,
        symmetric=True,
    ):
        if affine is not None:
            w_pos = apply_affine(affine, np.asarray(indices).T)

        valid_dirs = directions[indices]

        num_dirs = len(np.nonzero(np.abs(valid_dirs).max(axis=-1) > 0)[0])

        pnts_per_line = 2

        points_array = np.empty((num_dirs * pnts_per_line, 3))
        centers_array = np.empty_like(points_array, dtype=int)
        diffs_array = np.empty_like(points_array)
        line_count = 0
        for idx, center in enumerate(zip(indices[0], indices[1], indices[2])):
            if affine is None:
                xyz = np.asarray(center)
            else:
                xyz = w_pos[idx, :]
            valid_peaks = np.nonzero(np.abs(valid_dirs[idx, :, :]).max(axis=-1) > 0.0)[
                0
            ]
            for direction in valid_peaks:
                if values is not None:
                    pv = values[center][direction]
                else:
                    pv = 1.0

                if symmetric:
                    point_i = directions[center][direction] * pv + xyz
                    point_e = -directions[center][direction] * pv + xyz
                else:
                    point_i = directions[center][direction] * pv + xyz
                    point_e = xyz

                diff = point_e - point_i
                points_array[line_count * pnts_per_line, :] = point_e
                points_array[line_count * pnts_per_line + 1, :] = point_i
                centers_array[line_count * pnts_per_line, :] = center
                centers_array[line_count * pnts_per_line + 1, :] = center
                diffs_array[line_count * pnts_per_line, :] = diff
                diffs_array[line_count * pnts_per_line + 1, :] = diff
                line_count += 1

        vtk_points = numpy_to_vtk_points(points_array)

        vtk_cells = _points_to_vtk_cells(points_array)

        colors_tuple = _peaks_colors_from_points(points_array, colors=colors)
        vtk_colors, colors_are_scalars, self.__global_opacity = colors_tuple

        poly_data = PolyData()
        poly_data.SetPoints(vtk_points)
        poly_data.SetLines(vtk_cells)
        poly_data.GetPointData().SetScalars(vtk_colors)

        self.__mapper = PolyDataMapper()
        self.__mapper.SetInputData(poly_data)
        self.__mapper.ScalarVisibilityOn()
        self.__mapper.SetScalarModeToUsePointFieldData()
        self.__mapper.SelectColorArray("colors")
        self.__mapper.Update()

        self.SetMapper(self.__mapper)

        attribute_to_actor(self, centers_array, "center")
        attribute_to_actor(self, diffs_array, "diff")

        vs_var_dec = """
            in vec3 center;
            in vec3 diff;
            flat out vec3 centerVertexMCVSOutput;
            """
        fs_var_dec = """
            flat in vec3 centerVertexMCVSOutput;
            uniform bool isRange;
            uniform vec3 crossSection;
            uniform vec3 lowRanges;
            uniform vec3 highRanges;
            """
        orient_to_rgb = import_fury_shader(pjoin("utils", "orient_to_rgb.glsl"))
        visible_cross_section = import_fury_shader(
            pjoin("interaction", "visible_cross_section.glsl")
        )
        visible_range = import_fury_shader(pjoin("interaction", "visible_range.glsl"))

        vs_dec = compose_shader([vs_var_dec, orient_to_rgb])
        fs_dec = compose_shader([fs_var_dec, visible_cross_section, visible_range])

        vs_impl = """
            centerVertexMCVSOutput = center;
            if (vertexColorVSOutput.rgb == vec3(0))
            {
                vertexColorVSOutput.rgb = orient2rgb(diff);
            }
            """

        fs_impl = """
            if (isRange)
            {
                if (!inVisibleRange(centerVertexMCVSOutput))
                    discard;
            }
            else
            {
                if (!inVisibleCrossSection(centerVertexMCVSOutput))
                    discard;
            }
            """

        shader_to_actor(self, "vertex", decl_code=vs_dec, impl_code=vs_impl)
        shader_to_actor(self, "fragment", decl_code=fs_dec)
        shader_to_actor(self, "fragment", impl_code=fs_impl, block="light")

        # Color scale with a lookup table
        if colors_are_scalars:
            if lookup_colormap is None:
                lookup_colormap = colormap_lookup_table()

            self.__mapper.SetLookupTable(lookup_colormap)
            self.__mapper.UseLookupTableScalarRangeOn()
            self.__mapper.Update()

        self.__lw = linewidth
        self.GetProperty().SetLineWidth(self.__lw)

        if self.__global_opacity >= 0:
            self.GetProperty().SetOpacity(self.__global_opacity)

        self.__min_centers = np.zeros(shape=(3,))
        self.__max_centers = np.array(directions.shape[:3])

        self.__is_range = True
        self.__low_ranges = self.__min_centers
        self.__high_ranges = self.__max_centers
        self.__cross_section = self.__high_ranges // 2

        self.__mapper.AddObserver(
            Command.UpdateShaderEvent, self.__display_peaks_vtk_callback
        )

    @calldata_type(VTK_OBJECT)
    def __display_peaks_vtk_callback(self, caller, event, calldata=None):
        if calldata is not None:
            calldata.SetUniformi("isRange", self.__is_range)
            calldata.SetUniform3f("highRanges", self.__high_ranges)
            calldata.SetUniform3f("lowRanges", self.__low_ranges)
            calldata.SetUniform3f("crossSection", self.__cross_section)

    def display_cross_section(self, x, y, z):
        if self.__is_range:
            self.__is_range = False
        self.__cross_section = [x, y, z]

    def display_extent(self, x1, x2, y1, y2, z1, z2):
        if not self.__is_range:
            self.__is_range = True
        self.__low_ranges = [x1, y1, z1]
        self.__high_ranges = [x2, y2, z2]

    @property
    def cross_section(self):
        return self.__cross_section

    @property
    def global_opacity(self):
        return self.__global_opacity

    @global_opacity.setter
    def global_opacity(self, opacity):
        self.__global_opacity = opacity
        self.GetProperty().SetOpacity(self.__global_opacity)

    @property
    def high_ranges(self):
        return self.__high_ranges

    @property
    def is_range(self):
        return self.__is_range

    @property
    def low_ranges(self):
        return self.__low_ranges

    @property
    def linewidth(self):
        return self.__lw

    @linewidth.setter
    def linewidth(self, linewidth):
        self.__lw = linewidth
        self.GetProperty().SetLineWidth(self.__lw)

    @property
    def max_centers(self):
        return self.__max_centers

    @property
    def min_centers(self):
        return self.__min_centers


@warning_for_keywords()
def peak(
    peaks_dirs,
    *,
    peaks_values=None,
    mask=None,
    affine=None,
    colors=None,
    linewidth=1,
    lookup_colormap=None,
    symmetric=True,
):
    """Visualize peak directions as given from ``peaks_from_model`` function.

    Parameters
    ----------
    peaks_dirs : ndarray
        Peak directions. The shape of the array should be (X, Y, Z, D, 3).
    peaks_values : ndarray, optional
        Peak values. The shape of the array should be (X, Y, Z, D).
    affine : array, optional
        4x4 transformation array from native coordinates to world coordinates.
    mask : ndarray, optional
        3D mask
    colors : tuple or None, optional
        Default None. If None then every peak gets an orientation color
        in similarity to a DEC map.
    lookup_colormap : vtkLookupTable, optional
        Add a default lookup table to the colormap. Look at
        :func:`fury.actor.colormap_lookup_table` for more information.
    linewidth : float, optional
        Line thickness. Default is 1.
    symmetric : bool, optional
        If True, peaks are drawn for both peaks_dirs and -peaks_dirs. Else,
        peaks are only drawn for directions given by peaks_dirs. Default is
        True.

    Returns
    -------
    peak_actor : PeakActor
        Actor or LODActor representing the peaks directions and/or
        magnitudes.

    """
    if peaks_dirs.ndim != 5:
        raise ValueError(
            "Invalid peak directions. The shape of the structure "
            f"must be (XxYxZxDx3). Your data has {peaks_dirs.ndim} dimensions."
            ""
        )
    if peaks_dirs.shape[4] != 3:
        raise ValueError(
            "Invalid peak directions. The shape of the last "
            "dimension must be 3. Your data has a last dimension "
            f"of {peaks_dirs.shape[4]}."
        )

    dirs_shape = peaks_dirs.shape

    if peaks_values is not None:
        if peaks_values.ndim != 4:
            raise ValueError(
                "Invalid peak values. The shape of the structure "
                f"must be (XxYxZxD). Your data has {peaks_values.ndim} dimensions."
            )
        vals_shape = peaks_values.shape
        if vals_shape != dirs_shape[:4]:
            raise ValueError(
                "Invalid peak values. The shape of the values "
                "must coincide with the shape of the directions."
            )

    valid_mask = np.abs(peaks_dirs).max(axis=(-2, -1)) > 0
    if mask is not None:
        if mask.ndim != 3:
            warnings.warn(
                "Invalid mask. The mask must be a 3D array. The "
                f"passed mask has {mask.ndim} dimensions. Ignoring passed "
                "mask.",
                UserWarning,
                stacklevel=2,
            )
        elif mask.shape != dirs_shape[:3]:
            warnings.warn(
                "Invalid mask. The shape of the mask must coincide "
                "with the shape of the directions. Ignoring passed "
                "mask.",
                UserWarning,
                stacklevel=2,
            )
        else:
            valid_mask = np.logical_and(valid_mask, mask)
    indices = np.where(valid_mask)

    return PeakActor(
        peaks_dirs,
        indices,
        values=peaks_values,
        affine=affine,
        colors=colors,
        lookup_colormap=lookup_colormap,
        linewidth=linewidth,
        symmetric=symmetric,
    )


@warning_for_keywords()
def _peaks_colors_from_points(points, *, colors=None, points_per_line=2):
    """Return a VTK scalar array containing colors information for each one of
    the peaks according to the policy defined by the parameter colors.

    Parameters
    ----------
    points : (N, 3) array or ndarray
        points coordinates array.
    colors : None or string ('rgb_standard') or tuple (3D or 4D) or
             array/ndarray (N, 3 or 4) or array/ndarray (K, 3 or 4) or
             array/ndarray(N, ) or array/ndarray (K, )
        If None a standard orientation colormap is used for every line.
        If one tuple of color is used. Then all streamlines will have the same
        color.
        If an array (N, 3 or 4) is given, where N is equal to the number of
        points. Then every point is colored with a different RGB(A) color.
        If an array (K, 3 or 4) is given, where K is equal to the number of
        lines. Then every line is colored with a different RGB(A) color.
        If an array (N, ) is given, where N is the number of points then these
        are considered as the values to be used by the colormap.
        If an array (K,) is given, where K is the number of lines then these
        are considered as the values to be used by the colormap.
    points_per_line : int (1 or 2), optional
        number of points per peak direction.

    Returns
    -------
    color_array : vtkDataArray
        vtk scalar array with name 'colors'.
    colors_are_scalars : bool
        indicates whether or not the colors are scalars to be interpreted by a
        colormap.
    global_opacity : float
        returns 1 if the colors array doesn't contain opacity otherwise -1.

    """
    num_pnts = len(points)
    num_lines = num_pnts // points_per_line
    colors_are_scalars = False
    global_opacity = 1
    if colors is None or colors == "rgb_standard":
        # Automatic RGB colors
        colors = np.asarray((0, 0, 0))
        color_array = numpy_to_vtk_colors(np.tile(255 * colors, (num_pnts, 1)))
    elif type(colors) is tuple:
        global_opacity = 1 if len(colors) == 3 else -1
        colors = np.asarray(colors)
        color_array = numpy_to_vtk_colors(np.tile(255 * colors, (num_pnts, 1)))
    else:
        colors = np.asarray(colors)
        if len(colors) == num_lines:
            pnts_colors = np.repeat(colors, points_per_line, axis=0)
            if colors.ndim == 1:  # Scalar per line
                color_array = numpy_support.numpy_to_vtk(pnts_colors, deep=True)
                colors_are_scalars = True
            elif colors.ndim == 2:  # RGB(A) color per line
                global_opacity = 1 if colors.shape[1] == 3 else -1
                color_array = numpy_to_vtk_colors(255 * pnts_colors)
        elif len(colors) == num_pnts:
            if colors.ndim == 1:  # Scalar per point
                color_array = numpy_support.numpy_to_vtk(colors, deep=True)
                colors_are_scalars = True
            elif colors.ndim == 2:  # RGB(A) color per point
                global_opacity = 1 if colors.shape[1] == 3 else -1
                color_array = numpy_to_vtk_colors(255 * colors)

    color_array.SetName("colors")
    return color_array, colors_are_scalars, global_opacity


@warning_for_keywords()
def _points_to_vtk_cells(points, *, points_per_line=2):
    """Return the VTK cell array for the peaks given the set of points
    coordinates.

    Parameters
    ----------
    points : (N, 3) array or ndarray
        points coordinates array.
    points_per_line : int (1 or 2), optional
        number of points per peak direction.

    Returns
    -------
    cell_array : vtkCellArray
        connectivity + offset information.

    """
    num_pnts = len(points)
    num_cells = num_pnts // points_per_line

    cell_array = CellArray()

    """
    Connectivity is an array that contains the indices of the points that
    need to be connected in the visualization. The indices start from 0.
    """
    connectivity = np.asarray(list(range(0, num_pnts)), dtype=int)
    """
    Offset is an array that contains the indices of the first point of
    each line. The indices start from 0 and given the known geometry of
    this actor the creation of this array requires a 2 points padding
    between indices.
    """
    offset = np.asarray(list(range(0, num_pnts + 1, points_per_line)), dtype=int)

    vtk_array_type = numpy_support.get_vtk_array_type(connectivity.dtype)
    cell_array.SetData(
        numpy_support.numpy_to_vtk(offset, deep=True, array_type=vtk_array_type),
        numpy_support.numpy_to_vtk(connectivity, deep=True, array_type=vtk_array_type),
    )

    cell_array.SetNumberOfCells(num_cells)
    return cell_array


class PeaksVisualizer:
    def __init__(self, pam, world_coords):
        self._peak_dirs, self._affine = pam
        if world_coords:
            self._peak_actor = peak(self._peak_dirs, affine=self._affine)
        else:
            self._peak_actor = peak(self._peak_dirs)

    @property
    def actors(self):
        return [self._peak_actor]



import os
import tempfile

import numpy as np


def count_sketch(matrixa_name, matrixa_dtype, matrixa_shape, sketch_rows, tmp_dir):
    """Count Sketching algorithm to reduce the size of the matrix.

    Parameters
    ----------
    matrixa_name : str
        The name of the memmap file containing the matrix A.
    matrixa_dtype : dtype
        The dtype of the matrix A.
    matrixa_shape : tuple
        The shape of the matrix A.
    sketch_rows : int
        The number of rows in the sketch matrix.
    tmp_dir : str
        The directory to save the temporary files.

    Returns
    -------
    matrixc_file.name : str
        The name of the memmap file containing the sketch matrix.
    matrixc.dtype : dtype
        The dtype of the sketch matrix.
    matrixc.shape : tuple
        The shape of the sketch matrix.

    """
    matrixa = np.squeeze(
        np.memmap(matrixa_name, dtype=matrixa_dtype, mode="r+", shape=matrixa_shape)
    ).reshape(np.prod(matrixa_shape[:-1]), matrixa_shape[-1])

    with tempfile.NamedTemporaryFile(
        delete=False, dir=tmp_dir, suffix="matrix_t"
    ) as matrixt_file:
        matrixt = np.memmap(
            matrixt_file.name, dtype=matrixa_dtype, mode="w+", shape=matrixa.shape
        )
        hashed_indices = np.random.choice(sketch_rows, matrixa.shape[0], replace=True)
        rand_signs = np.random.choice(2, matrixa.shape[0], replace=True) * 2 - 1
        for i in range(0, matrixa.shape[0], matrixa.shape[0] // 20):
            end_index = min(i + matrixa.shape[0] // 20, matrixa.shape[0])
            matrixt[i:end_index, :] = (
                matrixa[i:end_index, :] * rand_signs[i:end_index, np.newaxis]
            )

    with tempfile.NamedTemporaryFile(
        delete=False, dir=tmp_dir, suffix="matrix_C"
    ) as matrixc_file:
        matrixc = np.memmap(
            matrixc_file.name,
            dtype=matrixa_dtype,
            mode="w+",
            shape=(sketch_rows, matrixa.shape[1]),
        )
        np.add.at(matrixc, hashed_indices, matrixt)
        matrixc.flush()
        matrixt.flush()
    del matrixt
    os.unlink(matrixt_file.name)
    return matrixc_file.name, matrixc.dtype, matrixc.shape



"""Learning algorithms for tractography"""

import numpy as np

import dipy.tracking.distances as pf


def detect_corresponding_tracks(indices, tracks1, tracks2):
    """Detect corresponding tracks from list tracks1 to list tracks2
    where tracks1 & tracks2 are lists of tracks

    Parameters
    ----------
    indices : sequence
       of indices of tracks1 that are to be detected in tracks2
    tracks1 : sequence
       of tracks as arrays, shape (N1,3) .. (Nm,3)
    tracks2 : sequence
       of tracks as arrays, shape (M1,3) .. (Mm,3)

    Returns
    -------
    track2track : array (N,2) where N is len(indices) of int
       it shows the correspondence in the following way:
       the first column is the current index in tracks1
       the second column is the corresponding index in tracks2

    Examples
    --------
    >>> import numpy as np
    >>> import dipy.tracking.learning as tl
    >>> A = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    >>> B = np.array([[1, 0, 0], [2, 0, 0], [3, 0, 0]])
    >>> C = np.array([[0, 0, -1], [0, 0, -2], [0, 0, -3]])
    >>> bundle1 = [A, B, C]
    >>> bundle2 = [B, A]
    >>> indices = [0, 1]
    >>> arr = tl.detect_corresponding_tracks(indices, bundle1, bundle2)

    Notes
    -----
    To find the corresponding tracks we use mam_distances with 'avg' option.
    Then we calculate the argmin of all the calculated distances and return it
    for every index. (See 3rd column of arr in the example given below.)


    """
    li = len(indices)

    track2track = np.zeros((li, 2))
    cnt = 0
    for i in indices:
        rt = [pf.mam_distances(tracks1[i], t, "avg") for t in tracks2]
        rt = np.array(rt)
        track2track[cnt] = np.array([i, rt.argmin()])
        cnt += 1

    return track2track.astype(int)


def detect_corresponding_tracks_plus(indices, tracks1, indices2, tracks2):
    """Detect corresponding tracks from 1 to 2 where tracks1 & tracks2 are
    sequences of tracks

    Parameters
    ----------
    indices : sequence
            of indices of tracks1 that are to be detected in tracks2
    tracks1 : sequence
            of tracks as arrays, shape (N1,3) .. (Nm,3)
    indices2 : sequence
            of indices of tracks2 in the initial brain
    tracks2 : sequence
            of tracks as arrays, shape (M1,3) .. (Mm,3)

    Returns
    -------
    track2track : array (N,2) where N is len(indices)
       of int showing the correspondence in th following way
       the first column is the current index of tracks1
       the second column is the corresponding index in tracks2

    Examples
    --------
    >>> import numpy as np
    >>> import dipy.tracking.learning as tl
    >>> A = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    >>> B = np.array([[1, 0, 0], [2, 0, 0], [3, 0, 0]])
    >>> C = np.array([[0, 0, -1], [0, 0, -2], [0, 0, -3]])
    >>> bundle1 = [A, B, C]
    >>> bundle2 = [B, A]
    >>> indices = [0, 1]
    >>> indices2 = indices
    >>> arr = tl.detect_corresponding_tracks_plus(indices, bundle1, indices2, bundle2)

    Notes
    -----
    To find the corresponding tracks we use mam_distances with 'avg' option.
    Then we calculate the argmin of all the calculated distances and return it
    for every index. (See 3rd column of arr in the example given below.)


    See Also
    --------
    distances.mam_distances

    """
    li = len(indices)
    track2track = np.zeros((li, 2))
    cnt = 0
    for i in indices:
        rt = [pf.mam_distances(tracks1[i], t, "avg") for t in tracks2]
        rt = np.array(rt)
        track2track[cnt] = np.array([i, indices2[rt.argmin()]])
        cnt += 1
    return track2track.astype(int)



# Tests for shore fitting
from math import factorial
import warnings

import numpy as np
import numpy.testing as npt
import pytest
from scipy.special import gamma, genlaguerre

from dipy.data import get_gtab_taiwan_dsi
from dipy.reconst.shm import descoteaux07_legacy_msg
from dipy.reconst.shore import ShoreModel
from dipy.sims.voxel import multi_tensor
from dipy.utils.optpkg import optional_package

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")
needs_cvxpy = pytest.mark.skipif(not have_cvxpy, reason="Requires CVXPY")


# Object to hold module global data
class _C:
    pass


data = _C()


def setup_module():
    data.gtab = get_gtab_taiwan_dsi()
    data.mevals = np.array(([0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]))
    data.angl = [(0, 0), (60, 0)]
    data.S, sticks = multi_tensor(
        data.gtab, data.mevals, S0=100.0, angles=data.angl, fractions=[50, 50], snr=None
    )
    data.radial_order = 6
    data.zeta = 700
    data.lambdaN = 1e-12
    data.lambdaL = 1e-12


def test_shore_error():
    data.gtab = get_gtab_taiwan_dsi()
    npt.assert_raises(ValueError, ShoreModel, data.gtab, radial_order=-4)
    npt.assert_raises(ValueError, ShoreModel, data.gtab, radial_order=7)
    npt.assert_raises(
        ValueError, ShoreModel, data.gtab, constrain_e0=False, positive_constraint=True
    )


@needs_cvxpy
def test_shore_positive_constrain():
    asm = ShoreModel(
        data.gtab,
        radial_order=data.radial_order,
        zeta=data.zeta,
        lambdaN=data.lambdaN,
        lambdaL=data.lambdaL,
        constrain_e0=True,
        positive_constraint=True,
        pos_grid=11,
        pos_radius=20e-03,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        asmfit = asm.fit(data.S)
        eap = asmfit.pdf_grid(11, 20e-03)
    npt.assert_almost_equal(eap[eap < 0].sum(), 0, 3)


def test_shore_fitting_no_constrain_e0():
    asm = ShoreModel(
        data.gtab,
        radial_order=data.radial_order,
        zeta=data.zeta,
        lambdaN=data.lambdaN,
        lambdaL=data.lambdaL,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        asmfit = asm.fit(data.S)
    npt.assert_almost_equal(compute_e0(asmfit), 1)


@needs_cvxpy
def test_shore_fitting_constrain_e0():
    asm = ShoreModel(
        data.gtab,
        radial_order=data.radial_order,
        zeta=data.zeta,
        lambdaN=data.lambdaN,
        lambdaL=data.lambdaL,
        constrain_e0=True,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        asmfit = asm.fit(data.S)
    npt.assert_almost_equal(compute_e0(asmfit), 1)


def compute_e0(shorefit):
    signal_0 = 0

    for n in range(int(shorefit.model.radial_order / 2) + 1):
        signal_0 += shorefit.shore_coeff[n] * (
            genlaguerre(n, 0.5)(0)
            * (
                (factorial(n))
                / (2 * np.pi * (shorefit.model.zeta**1.5) * gamma(n + 1.5))
            )
            ** 0.5
        )

    return signal_0



from abc import ABC, abstractmethod
from dataclasses import dataclass
import logging
from typing import Any
import warnings

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.util import show_ellipsis

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from fury import ui
    from fury.data import read_viz_icons


@dataclass
class HorizonUIElement:
    """Dataclass to define properties of horizon ui elements."""

    visibility: bool
    selected_value: Any
    obj: Any
    position = (0, 0)
    size = ("auto", "auto")


class HorizonTab(ABC):
    """Base for different tabs available in horizon."""

    def __init__(self):
        self._elements = []
        self.hide = lambda *args: None
        self.show = lambda *args: None

    @abstractmethod
    def build(self, tab_id):
        """Build all the elements under the tab.

        Parameters
        ----------
        tab_id : int
            Id of the tab.
        """

    def _register_elements(self, *args):
        """Register elements for rendering.

        Parameters
        ----------
        *args : HorizonUIElement(s)
            Elements to be register for rendering.
        """
        self._elements += list(args)

    def _toggle_actors(self, checkbox):
        """Toggle actors in the scene. This helps removing the actor to
        interact with actors behind this actor.

        Parameters
        ----------
        checkbox : CheckBox2D
        """
        if "" in checkbox.checked_labels:
            self.show(*self.actors)
        else:
            self.hide(*self.actors)

    def on_tab_selected(self):
        """Implement if require to update something while the tab becomes
        active.
        """
        if hasattr(self, "_actor_toggle"):
            self._toggle_actors(self._actor_toggle.obj)

    @property
    @abstractmethod
    def name(self):
        """Name of the tab."""

    @property
    @abstractmethod
    def actors(self):
        """Name of the tab."""

    @property
    def tab_id(self):
        """Id of the tab. Reference for Tab Manager to identify the tab.

        Returns
        -------
        int
        """
        return self._tab_id

    @property
    def elements(self):
        """list of underlying FURY ui elements in the tab."""
        return self._elements


class TabManager:
    """
    A Manager for tabs of the table panel.

    Attributes
    ----------

    tab_ui : TabUI
        Underlying FURY TabUI object.
    """

    @warning_for_keywords()
    def __init__(
        self,
        tabs,
        win_size,
        on_tab_changed,
        add_to_scene,
        *,
        remove_from_scene,
        sync_slices=False,
        sync_volumes=False,
        sync_peaks=False,
    ):
        num_tabs = len(tabs)
        self._tabs = tabs
        self._add_to_scene = add_to_scene
        self._remove_from_scene = remove_from_scene
        self._synchronize_slices = sync_slices
        self._synchronize_volumes = sync_volumes
        self._synchronize_peaks = sync_peaks

        win_width, _win_height = win_size

        self._tab_size = (1280, 240)
        x_pad = np.rint((win_width - self._tab_size[0]) / 2)

        self._active_tab_id = num_tabs - 1

        self._tab_ui = ui.TabUI(
            position=(x_pad, 5),
            size=self._tab_size,
            nb_tabs=num_tabs,
            active_color=(1, 1, 1),
            inactive_color=(0.5, 0.5, 0.5),
            draggable=True,
            startup_tab_id=self._active_tab_id,
        )

        self._tab_ui.on_change = self._tab_selected

        self.tab_changed = on_tab_changed

        slices_tabs = list(
            filter(lambda x: x.__class__.__name__ == "SlicesTab", self._tabs)
        )
        if not self._synchronize_slices and slices_tabs:
            msg = (
                "Images are of different dimensions, "
                + "synchronization of slices will not work"
            )
            logging.warning(msg)

        for tab_id, tab in enumerate(tabs):
            self._tab_ui.tabs[tab_id].title_font_size = 18
            tab.hide = self._hide_elements
            tab.show = self._show_elements
            tab.build(tab_id)
            if tab.__class__.__name__ == "SlicesTab":
                tab.on_volume_change = self.synchronize_volumes
            if tab.__class__.__name__ in ["SlicesTab", "PeaksTab"]:
                tab.on_slice_change = self.synchronize_slices
            self._render_tab_elements(tab.tab_id, tab.elements)

    def handle_text_overflows(self):
        for tab_id, tab in enumerate(self._tabs):
            self._handle_title_overflow(tab.name, self._tab_ui.tabs[tab_id].text_block)
            if tab.__class__.__name__ == "SlicesTab":
                self._handle_label_text_overflow(tab.elements)

    def _handle_label_text_overflow(self, elements):
        for element in elements:
            if (
                not element.size[0] == "auto"
                and element.obj.__class__.__name__ == "TextBlock2D"
                and isinstance(element.position, tuple)
            ):
                element.obj.message = show_ellipsis(
                    element.selected_value, element.obj.size[0], element.size[0]
                )

    def _handle_title_overflow(self, title_text, title_block):
        """Handle overflow of the tab title and show ellipsis if required.

        Parameters
        ----------
        title_text : str
            Text to be shown on the tab.
        title_block : TextBlock2D
            Fury UI element for holding the title of the tab.
        """
        tab_text = title_text.split(".", 1)[0]
        title_block.message = tab_text
        available_space, _ = self._tab_size
        text_size = title_block.size[0]
        max_width = (available_space / len(self._tabs)) - 15
        title_block.message = show_ellipsis(tab_text, text_size, max_width)

    def _render_tab_elements(self, tab_id, elements):
        for element in elements:
            if isinstance(element.position, list):
                for i, position in enumerate(element.position):
                    self._tab_ui.add_element(tab_id, element.obj[i], position)
            else:
                self._tab_ui.add_element(tab_id, element.obj, element.position)

    def _hide_elements(self, *args):
        """Hide elements from the scene.

        Parameters
        ----------
        *args : HorizonUIElement or FURY actors
            Elements to be hidden.
        """
        self._remove_from_scene(*self._get_vtkActors(*args))

    def _show_elements(self, *args):
        """Show elements in the scene.

        Parameters
        ----------
        *args : HorizonUIElement or FURY actors
            Elements to be hidden.
        """
        self._add_to_scene(*self._get_vtkActors(*args))

    def _get_vtkActors(self, *args):
        elements = []
        vtk_actors = []
        for element in args:
            if element.__class__.__name__ == "HorizonUIElement":
                if isinstance(element.obj, list):
                    for obj in element.obj:
                        elements.append(obj)
                else:
                    elements.append(element.obj)
            else:
                elements.append(element)

        for element in elements:
            if hasattr(element, "_get_actors") and callable(element._get_actors):
                vtk_actors += element.actors
            else:
                vtk_actors.append(element)
        return vtk_actors

    def _tab_selected(self, tab_ui):
        if self._active_tab_id == tab_ui.active_tab_idx:
            self._active_tab_id = -1
            return

        self._active_tab_id = tab_ui.active_tab_idx

        current_tab = self._tabs[self._active_tab_id]
        self.tab_changed(current_tab.actors)
        current_tab.on_tab_selected()

    def reposition(self, win_size):
        """
        Reposition the tabs panel.

        Parameters
        ----------
        win_size : (float, float)
            size of the horizon window.
        """
        win_width, _win_height = win_size
        x_pad = np.rint((win_width - self._tab_size[0]) / 2)
        self._tab_ui.position = (x_pad, 5)

    def synchronize_slices(self, active_tab_id, x_value, y_value, z_value):
        """
        Synchronize slicers for all the images and peaks.

        Parameters
        ----------
        active_tab_id: int
            tab_id of the action performing tab
        x_value: float
            x-value of the active slicer
        y_value: float
            y-value of the active slicer
        z_value: float
            z-value of the active slicer
        """

        if not self._synchronize_slices and not self._synchronize_peaks:
            return

        for tab in self._get_non_active_tabs(active_tab_id, ["SlicesTab", "PeaksTab"]):
            tab.update_slices(x_value, y_value, z_value)

    def synchronize_volumes(self, active_tab_id, value):
        """Synchronize volumes for all the images with volumes.

        Parameters
        ----------
        active_tab_id : int
            tab_id of the action performing tab
        value : float
            volume value of the active volume slider

        """

        if not self._synchronize_volumes:
            return

        for slices_tab in self._get_non_active_tabs(active_tab_id):
            slices_tab.update_volume(value)

    def _get_non_active_tabs(self, active_tab_id, types=("SlicesTab",)):
        """Get tabs which are not active and slice tabs.

        Parameters
        ----------
        active_tab_id : int
        types : list(str), optional

        Returns
        -------
        list
        """
        return list(
            filter(
                lambda x: x.__class__.__name__ in types
                and not x.tab_id == active_tab_id,
                self._tabs,
            )
        )

    @property
    def tab_ui(self):
        """FURY TabUI object."""
        return self._tab_ui


@warning_for_keywords()
def build_label(text, *, font_size=16, bold=False):
    """Simple utility function to build labels.

    Parameters
    ----------
    text : str
    font_size : int, optional
    bold : bool, optional

    Returns
    -------
    label : TextBlock2D
    """

    label = ui.TextBlock2D()
    label.message = text
    label.font_size = font_size
    label.font_family = "Arial"
    label.justification = "left"
    label.bold = bold
    label.italic = False
    label.shadow = False
    label.actor.GetTextProperty().SetBackgroundColor(0, 0, 0)
    label.actor.GetTextProperty().SetBackgroundOpacity(0.0)
    label.color = (0.7, 0.7, 0.7)

    return HorizonUIElement(True, text, label)


@warning_for_keywords()
def build_slider(
    initial_value,
    max_value,
    *,
    min_value=0,
    length=450,
    line_width=3,
    radius=8,
    font_size=16,
    text_template="{value:.1f} ({ratio:.0%})",
    on_moving_slider=lambda _slider: None,
    on_value_changed=lambda _slider: None,
    on_change=lambda _slider: None,
    on_handle_released=lambda _istyle, _obj, _slider: None,
    label="",
    label_font_size=16,
    label_style_bold=False,
    is_double_slider=False,
):
    """Create a horizon theme based disk-knob slider.

    Parameters
    ----------
    initial_value : float, (float, float)
        Initial value(s) of the slider.
    max_value : float
        Maximum value of the slider.
    min_value : float, optional
        Minimum value of the slider.
    length : int, optional
        Length of the slider.
    line_width : int, optional
        Width of the line on which the disk will slide.
    radius : int, optional
        Radius of the disk handle.
    font_size : int, optional
        Size of the text to display alongside the slider (pt).
    text_template : str, callable, optional
        If str, text template can contain one or multiple of the
        replacement fields: `{value:}`, `{ratio:}`.
        If callable, this instance of `:class:LineSlider2D` will be
        passed as argument to the text template function.
    on_moving_slider : callable, optional
        When the slider is interacted by the user.
    on_value_changed : callable, optional
        When value of the slider changed programmatically.
    on_change : callable, optional
        When value of the slider changed.
    on_handle_released: callable, optional
        When handle released.
    label : str, optional
        Label to ui element for slider
    label_font_size : int, optional
        Size of label text to display with slider
    label_style_bold : bool, optional
        Is label should have bold style.
    is_double_slider : bool, optional
        True if the slider allows to adjust two values.

    Returns
    -------
    label : HorizonUIElement
        Slider label.
    HorizonUIElement
        Slider.
    """

    if is_double_slider and "ratio" in text_template:
        warnings.warn("Double slider only support values and not ratio", stacklevel=2)
        return

    slider_label = build_label(label, font_size=label_font_size, bold=label_style_bold)

    if not is_double_slider:
        slider = ui.LineSlider2D(
            initial_value=initial_value,
            max_value=max_value,
            min_value=min_value,
            length=length,
            line_width=line_width,
            outer_radius=radius,
            font_size=font_size,
            text_template=text_template,
        )
    else:
        slider = ui.LineDoubleSlider2D(
            initial_values=initial_value,
            max_value=max_value,
            min_value=min_value,
            length=length,
            line_width=line_width,
            outer_radius=radius,
            font_size=font_size,
            text_template=text_template,
        )

    slider.on_moving_slider = on_moving_slider
    slider.on_value_changed = on_value_changed
    slider.on_change = on_change

    if not is_double_slider:
        slider.handle_events(slider.handle.actor)
        slider.on_left_mouse_button_released = on_handle_released

    slider.default_color = (1.0, 0.5, 0.0)
    slider.track.color = (0.8, 0.3, 0.0)
    slider.active_color = (0.9, 0.4, 0.0)
    if not is_double_slider:
        slider.handle.color = (1.0, 0.5, 0.0)
    else:
        slider.handles[0].color = (1.0, 0.5, 0.0)
        slider.handles[1].color = (1.0, 0.5, 0.0)

    return slider_label, HorizonUIElement(True, initial_value, slider)


@warning_for_keywords()
def build_checkbox(
    *,
    labels=None,
    checked_labels=None,
    padding=1,
    font_size=16,
    on_change=lambda _checkbox: None,
):
    """Create horizon theme checkboxes.

    Parameters
    ----------
    labels : list(str), optional
        List of labels of each option.
    checked_labels: list(str), optional
        List of labels that are checked on setting up.
    padding : float, optional
        The distance between two adjacent options element
    font_size : int, optional
        Size of the text font.
    on_change : callback, optional
        When checkbox value changed

    Returns
    -------
    checkbox : HorizonUIElement
    """

    if labels is None or not labels:
        warnings.warn(
            "At least one label needs to be to create checkboxes", stacklevel=2
        )
        return

    if checked_labels is None:
        checked_labels = ()

    checkboxes = ui.Checkbox(
        labels=labels,
        checked_labels=checked_labels,
        padding=padding,
        font_size=font_size,
    )

    checkboxes.on_change = on_change

    return HorizonUIElement(True, checked_labels, checkboxes)


@warning_for_keywords()
def build_radio_button(
    *,
    labels=None,
    checked_labels=None,
    padding=1,
    font_size=16,
    on_change=lambda _checkbox: None,
):
    """Create horizon theme radio buttons.

    Parameters
    ----------
    labels : list(str), optional
        List of labels of each option.
    checked_labels: list(str), optional
        List of labels that are checked on setting up.
    padding : float, optional
        The distance between two adjacent options element
    font_size : int, optional
        Size of the text font.
    on_change : callback, optional
        When radio button value changed

    Returns
    -------
    radio : HorizonUIElement
    """

    if labels is None or not labels:
        warnings.warn(
            "At least one label needs to be to create radio buttons", stacklevel=2
        )
        return

    if checked_labels is None:
        checked_labels = ()

    radio = ui.RadioButton(
        labels=labels,
        checked_labels=checked_labels,
        padding=padding,
        font_size=font_size,
    )

    radio.on_change = on_change

    return HorizonUIElement(True, checked_labels, radio)


@warning_for_keywords()
def build_switcher(
    *,
    items=None,
    label="",
    initial_selection=0,
    on_prev_clicked=lambda _selected_value: None,
    on_next_clicked=lambda _selected_value: None,
    on_value_changed=lambda _selected_idx, _selected_value: None,
):
    """Create horizon theme switcher.

    Parameters
    ----------
    items : list, optional
        dictionaries with keys 'label' and 'value'. Label will be used to show
        it to user and value will be used for selection.
    label : str, optional
        label for the switcher.
    initial_selection : int, optional
        index of the selected item initially.
    on_prev_clicked : callback, optional
        method providing a callback when prev value is selected in switcher.
    on_next_clicked : callback, optional
        method providing a callback when next value is selected in switcher.
    on_value_changed : callback, optional
        method providing a callback when either prev or next value selected in
        switcher.

    Returns
    -------
    HorizonCombineElement(
        label: HorizonUIElement,
        element(switcher): HorizonUIElement)

    Notes
    -----
    switcher: consists 'obj' which is an array providing FURY UI elements used.
    """
    if items is None:
        warnings.warn("No items passed in switcher", stacklevel=2)
        return

    num_items = len(items)

    if initial_selection >= num_items:
        initial_selection = 0

    switch_label = build_label(text=label)
    selection_label = build_label(text=items[initial_selection]["label"]).obj

    left_button = ui.Button2D(
        icon_fnames=[("left", read_viz_icons(fname="circle-left.png"))], size=(25, 25)
    )
    right_button = ui.Button2D(
        icon_fnames=[("right", read_viz_icons(fname="circle-right.png"))], size=(25, 25)
    )

    switcher = HorizonUIElement(
        True,
        [initial_selection, items[initial_selection]["value"]],
        [left_button, selection_label, right_button],
    )

    def left_clicked(_i_ren, _obj, _button):
        selected_id = switcher.selected_value[0] - 1
        if selected_id < 0:
            selected_id = num_items - 1
        value_changed(selected_id)
        on_prev_clicked(items[selected_id]["value"])
        on_value_changed(selected_id, items[selected_id]["value"])

    def right_clicked(_i_ren, _obj, _button):
        selected_id = switcher.selected_value[0] + 1
        if selected_id >= num_items:
            selected_id = 0
        value_changed(selected_id)
        on_next_clicked(items[selected_id]["value"])
        on_value_changed(selected_id, items[selected_id]["value"])

    def value_changed(selected_id):
        switcher.selected_value[0] = selected_id
        switcher.selected_value[1] = items[selected_id]["value"]
        selection_label.message = items[selected_id]["label"]

    left_button.on_left_mouse_button_clicked = left_clicked
    right_button.on_left_mouse_button_clicked = right_clicked

    return (switch_label, switcher)



"""Routines to support optional packages"""

import importlib

from packaging.version import Version

try:
    import pytest
except ImportError:
    have_pytest = False
else:
    have_pytest = True

from dipy.utils.tripwire import TripWire


def optional_package(name, *, trip_msg=None, min_version=None):
    """Return package-like thing and module setup for package `name`

    Parameters
    ----------
    name : str
        package name
    trip_msg : None or str
        message to give when someone tries to use the return package, but we
        could not import it, and have returned a TripWire object instead.
        Default message if None.
    min_version : None or str
        If not None, require that the imported package be at least this
        version.  If the package has no ``__version__`` attribute, or if the
        version is not parseable, raise an error.

    Returns
    -------
    pkg_like : module or ``TripWire`` instance
        If we can import the package, return it.  Otherwise return an object
        raising an error when accessed
    have_pkg : bool
        True if import for package was successful, false otherwise
    module_setup : function
        callable usually set as ``setup_module`` in calling namespace, to allow
        skipping tests.

    Examples
    --------
    Typical use would be something like this at the top of a module using an
    optional package:

    >>> from dipy.utils.optpkg import optional_package
    >>> pkg, have_pkg, setup_module = optional_package('not_a_package')

    Of course in this case the package doesn't exist, and so, in the module:

    >>> have_pkg
    False

    and

    >>> pkg.some_function() #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TripWireError: We need package not_a_package for these functions, but
    ``import not_a_package`` raised an ImportError

    If the module does exist - we get the module

    >>> pkg, _, _ = optional_package('os')
    >>> hasattr(pkg, 'path')
    True

    Or a submodule if that's what we asked for

    >>> subpkg, _, _ = optional_package('os.path')
    >>> hasattr(subpkg, 'dirname')
    True
    """
    try:
        pkg = importlib.import_module(name)
    except ImportError:
        pass
    else:  # import worked
        # top level module
        if not min_version:
            return pkg, True, lambda: None

        current_version = getattr(pkg, "__version__", "0.0.0")
        if Version(current_version) >= Version(min_version):
            return pkg, True, lambda: None

        if trip_msg is None:
            trip_msg = (
                f"We need at least version {min_version} of "
                f"package {name}, but ``import {name}`` "
                f"found version {current_version}."
            )
            if current_version == "0.0.0":
                trip_msg += "Your installation might be incomplete or corrupted."

    if trip_msg is None:
        trip_msg = (
            f"We need package {name} for these functions, but "
            f"``import {name}`` raised an ImportError"
        )
    pkg = TripWire(trip_msg)

    def setup_module():
        if have_pytest:
            pytest.mark.skip(f"No {name} for these tests")

    return pkg, False, setup_module



"""
============================================================
Creating a new workflow.
============================================================

A workflow is a series of DIPY_ operations with fixed inputs and outputs
that is callable via command line or another interface.

For example, after installing DIPY_, you can call anywhere from your command
line::

    dipy_nlmeans t1.nii.gz t1_denoised.nii.gz

First create your workflow (let's name this workflow file as my_workflow.py).
Usually this is a python file in the ``<../dipy/workflows>`` directory.
"""

import shutil

###############################################################################
# ``shutil`` Will be used for sample file manipulation.
from dipy.workflows.workflow import Workflow

###############################################################################
# ``Workflow`` is the base class that will be extended to create our workflow.


class AppendTextFlow(Workflow):
    def run(
        self, input_files, text_to_append="dipy", out_dir="", out_file="append.txt"
    ):
        """
        Parameters
        ----------
        input_files : string
            Path to the input files. This path may contain wildcards to
            process multiple inputs at once.

        text_to_append : string, optional
            Text that will be appended to the file. (default 'dipy')

        out_dir : string, optional
            Where the resulting file will be saved. (default '')

        out_file : string, optional
            Name of the result file to be saved. (default 'append.txt')
        """

        """
        ``AppendTextFlow`` is the name of our workflow. Note that it needs
        to extend Workflow for everything to work properly. It will append
        text to a file.

        It is mandatory to have out_dir as a parameter. It is also mandatory
        to put `out_` in front of every parameter that is going to be an
        output. Lastly, all `out_` params needs to be at the end of the params
        list.

        The ``run`` docstring is very important, you need to document every
        parameter as they will be used with inspection to build the command line
        argument parser.
        """

        io_it = self.get_io_iterator()

        for in_file, out_file in io_it:
            shutil.copy(in_file, out_file)

            with open(out_file, "a") as myfile:
                myfile.write(text_to_append)


###############################################################################
# Use self.get_io_iterator() in every workflow you create. This creates
# an ``IOIterator`` object that create output file names and directory
# structure based on the inputs and some other advanced output strategy
# parameters.
#
# By iterating on the ``IOIterator`` object you created previously you
# conveniently get all input and output paths for every input file
# found when globbing the input parameters.
#
# The code in the loop is the actual workflow processing code. It can be
# anything. For example, it just appends text to an input file.
#
# This is it for the workflow! Now to be able to call it easily via command
# line, you need to add this workflow in 2 different files:
#
# - ``<dipy_root>/pyproject.toml``: open this file and add the following line
#   to the ``[project.scripts]`` section:
#   ``dipy_append_text = "dipy.workflows.cli:run"``
# - ``<dipy_root>/dipy/workflows/cli.py``: open this file and add the workflow
#   information to the ``cli_flows`` dictionary. The key is the name of the
#   command line command and the value is a tuple with the module name and the
#   workflow class name. In this case it would be:
#   ``"dipy_append_text": ("dipy.workflows.my_workflow", "AppendTextFlow")``
#
# That`s it! Now you can call your workflow from anywhere with the command line.
# Let's just call the script you just made with ``-h`` to see the argparser help
# text::
#
#    dipy_append_text --help
#
# You should see all your parameters available along with some extra common
# ones like logging file and force overwrite. Also all the documentation you
# wrote about each parameter is there.



import math

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_raises

from dipy.core.gradients import gradient_table
from dipy.core.sphere import HemiSphere, disperse_charges
import dipy.reconst.cti as cti
from dipy.reconst.cti import (
    from_qte_to_cti,
    ls_fit_cti,
    multi_gaussian_k_from_c,
    split_cti_params,
)
from dipy.reconst.dki import (
    axial_kurtosis,
    kurtosis_fractional_anisotropy,
    mean_kurtosis,
    mean_kurtosis_tensor,
    radial_kurtosis,
)
from dipy.reconst.dti import decompose_tensor, mean_diffusivity
import dipy.reconst.qti as qti
from dipy.reconst.tests.test_qti import _anisotropic_DTD, _isotropic_DTD
from dipy.reconst.utils import cti_design_matrix as design_matrix

gtab1, gtab2, gtab, DTDs, S0 = None, None, None, None, None


def setup_module():
    global gtab1, gtab2, gtab, DTDs, S0
    # Generating the DDE acquisition parameters (gtab1 and gtab2) for CTI
    # based on the minimal requirements.
    # This code then further generates the corresponding QTE gtab to test
    # simulated signals based on multiple Gaussian components.
    # In this scenario, CTI should yield analogous Kaniso and Kiso values as
    # compared to QTE, while Kmicro would be zero.

    rng = np.random.default_rng(1234)
    n_pts = 20
    theta = np.pi * rng.random(n_pts)
    phi = 2 * np.pi * rng.random(n_pts)
    hsph_initial = HemiSphere(theta=theta, phi=phi)
    hsph_updated, potential = disperse_charges(hsph_initial, 5000)

    # Generating gtab1
    bvecs1 = np.concatenate([hsph_updated.vertices] * 4)
    bvecs1 = np.append(bvecs1, [[0, 0, 0]], axis=0)
    bvals1 = np.array([2] * 20 + [1] * 20 + [1] * 20 + [1] * 20 + [0])
    gtab1 = gradient_table(bvals1, bvecs=bvecs1)

    # Generating perpendicular directions to hsph_updated
    hsph_updated90 = _perpendicular_directions_temp(hsph_updated.vertices)
    dot_product = np.sum(hsph_updated.vertices * hsph_updated90, axis=1)
    assert all(np.isclose(dot_product, 0))

    # Generating gtab2
    bvecs2 = np.concatenate(
        ([hsph_updated.vertices] * 2) + [hsph_updated90] + ([hsph_updated.vertices])
    )
    bvecs2 = np.append(bvecs2, [[0, 0, 0]], axis=0)
    bvals2 = np.array([0] * 20 + [1] * 20 + [1] * 20 + [0] * 20 + [0])
    gtab2 = gradient_table(bvals2, bvecs=bvecs2)

    e1 = bvecs1
    e2 = bvecs2
    e3 = np.cross(e1, e2)
    V = np.stack((e1, e2, e3), axis=-1)
    V_transpose = np.transpose(V, axes=(0, 2, 1))
    B = np.zeros((81, 3, 3))
    b = np.zeros((3, 3))
    for i in range(81):
        b[0, 0] = bvals1[i]
        b[1, 1] = bvals2[i]
        B[i] = np.matmul(V[i], np.matmul(b, V_transpose[i]))
    gtab = gradient_table(bvals1, bvecs=bvecs1, btens=B)
    S0 = 100
    anisotropic_DTD = _anisotropic_DTD()
    isotropic_DTD = _isotropic_DTD()

    DTDs = [
        anisotropic_DTD,
        isotropic_DTD,
        np.concatenate((anisotropic_DTD, isotropic_DTD)),
    ]
    # DTD_labels = ["Anisotropic DTD", "Isotropic DTD", "Combined DTD"]


def teardown_module():
    global gtab1, gtab2, gtab, DTDs, S0
    gtab1, gtab2, gtab, DTDs, S0 = None, None, None, None, None


def _perpendicular_directions_temp(v, num=20, half=False):
    """
    Computes a set of perpendicular directions relative to the direction in v.

    The perpendicular directions are computed by sampling on a unit
    circumference that is perpendicular to `v`. The computation depends on
    whether the vector is aligned with the x-axis or not.

    Parameters
    ----------
    v : array (3,)
        Array containing the three cartesian coordinates of vector v.
    num : int, optional
        Number of perpendicular directions to generate. Default is 20.
    half : bool, optional
        If True, perpendicular directions are sampled on half of the unit
        circumference perpendicular to v, otherwise they are sampled on the
        full circumference. Default is False.

    Returns
    -------
    psamples : array (n, 3)
        Array of vectors perpendicular to v.
    """
    v = np.array(v, dtype=np.float64)
    v = v.T
    er = np.finfo(v[0].dtype).eps * 1e3
    if half is True:
        a = np.linspace(0.0, math.pi, num=num, endpoint=False)
    else:
        a = np.linspace(0.0, 2 * math.pi, num=num, endpoint=False)
    cosa = np.cos(a)
    sina = np.sin(a)
    if np.any(abs(v[0] - 1.0) > er):
        sq = np.sqrt(v[1] ** 2 + v[2] ** 2)
        psamples = np.array(
            [
                -sq * sina,
                (v[0] * v[1] * sina - v[2] * cosa) / sq,
                (v[0] * v[2] * sina + v[1] * cosa) / sq,
            ]
        )
    else:
        sq = np.sqrt(v[0] ** 2 + v[2] ** 2)
        psamples = np.array(
            [
                -(v[2] * cosa + v[0] * v[1] * sina) / sq,
                sina * sq,
                (v[0] * cosa - v[2] * v[1] * sina) / sq,
            ]
        )
    return psamples.T


def construct_cti_params(evals, evecs, kt, fct):
    """
    Combines all components to generate Correlation Tensor Model Parameter.

    Parameters
    ----------
    evals : array (..., 3)
        Eigenvalues from eigen decomposition of the tensor.
    evecs : array (..., 3)
        Associated eigenvectors from eigen decomposition of the tensor.
        Eigenvectors are columnar (e.g. evecs[:,j] is associated with
        evals[j])
    kt : array (..., 15)
        Fifteen elements of the kurtosis tensor
    fct: array(..., 21)
        Twenty-one elements of the covariance tensor

    Returns
    -------
    cti_params :  numpy.ndarray (..., 48)
    All parameters estimated from the correlation tensor model.
    Parameters are ordered as follows:

        1. Three diffusion tensor's eigenvalues
        2. Three lines of the eigenvector matrix each containing the
        first, second and third coordinates of the eigenvector
        3. Fifteen elements of the kurtosis tensor
        4. Twenty-One elements of the covariance tensor
    """
    fevals = evals.reshape((-1, evals.shape[-1]))
    fevecs = evecs.reshape((-1,) + evecs.shape[-2:])
    fevecs = fevecs.reshape((1, -1))
    fkt = kt.reshape((-1, kt.shape[-1]))
    cti_params = np.concatenate((fevals.T, fevecs.T, fkt, fct), axis=0)
    return np.squeeze(cti_params)


def test_cti_prediction():
    ctiM = cti.CorrelationTensorModel(gtab1, gtab2)
    for DTD in DTDs:
        D = np.mean(DTD, axis=0)
        evals, evecs = decompose_tensor(D)
        C = qti.dtd_covariance(DTD)
        C = qti.from_6x6_to_21x1(C)
        ccti = from_qte_to_cti(C)
        MD = mean_diffusivity(evals)
        K = multi_gaussian_k_from_c(ccti, MD)
        cti_params = construct_cti_params(evals, evecs, K, ccti)
        cti_pred_signals = ctiM.predict(cti_params, S0=S0)
        qti_pred_signals = qti.qti_signal(gtab, D, C, S0=S0)[np.newaxis, :]
        assert np.allclose(
            cti_pred_signals, qti_pred_signals
        ), "CTI and QTI signals do not match!"

        # check the function predict of the CorrelationTensorFit object
        ctiF = ctiM.fit(cti_pred_signals)
        ctiF_pred = ctiF.predict(gtab1, gtab2, S0=S0)
        assert_array_almost_equal(ctiF_pred, cti_pred_signals)


def test_split_cti_param():
    ctiM = cti.CorrelationTensorModel(gtab1, gtab2)
    for DTD in DTDs:
        D = np.mean(DTD, axis=0)
        evals, evecs = decompose_tensor(D)
        C = qti.dtd_covariance(DTD)
        C = qti.from_6x6_to_21x1(C)
        ccti = from_qte_to_cti(C)

        MD = mean_diffusivity(evals)
        K = multi_gaussian_k_from_c(ccti, MD)

        cti_params = construct_cti_params(evals, evecs, K, ccti)
        ctiM = cti.CorrelationTensorModel(gtab1, gtab2)
        cti_pred_signals = ctiM.predict(cti_params, S0=S0)
        ctiF = ctiM.fit(cti_pred_signals)
        evals, evecs, kt, ct = cti.split_cti_params(ctiF.model_params)

        assert_array_almost_equal(evals, ctiF.evals)
        assert_array_almost_equal(evecs, ctiF.evecs)
        assert np.allclose(kt, ctiF.kt), "kt doesn't match in test_split_cti_param "
        assert np.allclose(ct, ctiF.ct), "ct doesn't match in test_split_cti_param"


def test_cti_fits():
    ctiM = cti.CorrelationTensorModel(gtab1, gtab2)
    CTI_data = np.zeros((2, 2, 1, len(gtab1.bvals)))
    for DTD in DTDs:
        D = np.mean(DTD, axis=0)
        evals, evecs = decompose_tensor(D)
        C = qti.dtd_covariance(DTD)
        C = qti.from_6x6_to_21x1(C)
        ccti = from_qte_to_cti(C)
        MD = mean_diffusivity(evals)
        K = multi_gaussian_k_from_c(ccti, MD)
        cti_params = construct_cti_params(evals, evecs, K, ccti)
        cti_pred_signals = ctiM.predict(cti_params, S0=S0)
        evals, evecs, kt, ct = split_cti_params(cti_params)

        # Testing the model with correct min_signal value
        ctiM = cti.CorrelationTensorModel(gtab1, gtab2, min_signal=1)
        cti_pred_signals = ctiM.predict(cti_params, S0=S0)
        ctiF = ctiM.fit(cti_pred_signals)
        evals, evecs, kt, ct = cti.split_cti_params(ctiF.model_params)
        assert_array_almost_equal(evals, ctiF.evals)
        assert_array_almost_equal(evecs, ctiF.evecs)
        assert np.allclose(kt, ctiF.kt), "kt doesn't match in test_split_cti_param "
        assert np.allclose(ct, ctiF.ct), "ct doesn't match in test_split_cti_param"

        # Testing Multi-Voxel Fit
        CTI_data[0, 0, 0] = cti_pred_signals
        CTI_data[0, 1, 0] = cti_pred_signals
        CTI_data[1, 0, 0] = cti_pred_signals
        CTI_data[1, 1, 0] = cti_pred_signals
        multi_params = np.zeros((2, 2, 1, 48))
        multi_params[0, 0, 0] = multi_params[0, 1, 0] = cti_params
        multi_params[1, 0, 0] = multi_params[1, 1, 0] = cti_params
        ctiF_multi = ctiM.fit(CTI_data)

        multi_evals, _, multi_kt, multi_ct = split_cti_params(ctiF_multi.model_params)
        assert np.allclose(evals, multi_evals), "Evals don't match"
        assert np.allclose(kt, multi_kt), "K doesn't match"
        assert np.allclose(ct, multi_ct), "C doesn't match"

        # Check that it works with more than one voxel, and with a different S0
        # in each voxel:
        cti_multi_pred_signals = ctiM.predict(
            multi_params, S0=100 * np.ones(ctiF_multi.shape[:3])
        )
        CTI_data = cti_multi_pred_signals
        ctiF_multi_pred_signals = ctiM.fit(CTI_data)
        multi_evals, _, multi_kt, multi_ct = split_cti_params(
            ctiF_multi_pred_signals.model_params
        )
        assert np.allclose(evals, multi_evals), "Evals don't match"
        assert np.allclose(kt, multi_kt), "K doesn't match"
        assert np.allclose(ct, multi_ct), "C doesn't match"

        # Testing ls_fit_cti
        inverse_design_matrix = np.linalg.pinv(design_matrix(gtab1, gtab2))
        cti_return = ls_fit_cti(
            design_matrix(gtab1, gtab2), cti_pred_signals, inverse_design_matrix
        )
        evals_return, _, kt_return, ct_return = split_cti_params(cti_return)
        assert np.allclose(evals, evals_return), "evals do not match!"
        assert np.allclose(kt, kt_return), "K do not match!"
        assert np.allclose(ct, ct_return), "C do not match!"

        # OLS fitting
        ctiM = cti.CorrelationTensorModel(gtab1, gtab2, fit_method="OLS")
        ctiF = ctiM.fit(cti_pred_signals)
        ols_evals, _, ols_kt, ols_ct = split_cti_params(ctiF.model_params)
        assert np.allclose(evals, ols_evals), "evals do not match!"
        assert np.allclose(kt, ols_kt), "K do not match!"
        assert np.allclose(ct, ols_ct), "C do not match!"

        # WLS fitting
        cti_wlsM = cti.CorrelationTensorModel(gtab1, gtab2, fit_method="WLS")
        cti_wlsF = cti_wlsM.fit(cti_pred_signals)
        wls_evals, _, wls_kt, wls_ct = split_cti_params(cti_wlsF.model_params)
        assert np.allclose(evals, wls_evals), "evals do not match!"
        assert np.allclose(kt, wls_kt), "K do not match!"
        assert np.allclose(ct, wls_ct), "C do not match!"

        # checking Mean Kurtosis Values
        mk_result = ctiF.mk(min_kurtosis=-3.0 / 7, max_kurtosis=10, analytical=True)
        mean_kurtosis_result = mean_kurtosis(
            cti_params, min_kurtosis=-3.0 / 7, max_kurtosis=10, analytical=True
        )
        assert np.allclose(mk_result, mean_kurtosis_result), (
            "The results of the mk function from CorrelationTensorFit and the "
            "mean_kurtosis function from dki.py are not equal."
        )
        # checking Axial Kurtosis Values
        ak_result = ctiF.ak(min_kurtosis=-3.0 / 7, max_kurtosis=10, analytical=True)
        axial_kurtosis_result = axial_kurtosis(
            cti_params, min_kurtosis=-3.0 / 7, max_kurtosis=10, analytical=True
        )

        assert np.allclose(ak_result, axial_kurtosis_result), (
            "The results of the ak function from CorrelationTensorFit and the "
            "axial_kurtosis function from dki.py are not equal."
        )
        # checking Radial kurtosis values
        rk_result = ctiF.rk(min_kurtosis=-3.0 / 7, max_kurtosis=10, analytical=True)
        radial_kurtosis_result = radial_kurtosis(
            cti_params, min_kurtosis=-3.0 / 7, max_kurtosis=10, analytical=True
        )

        assert np.allclose(rk_result, radial_kurtosis_result), (
            "The results of the rk function from CorrelationTensorFit and the "
            "radial_kurtosis function from DKI.py are not equal."
        )
        # checking Anisotropic values.
        kfa_result = ctiF.kfa
        kurtosis_fractional_anisotropy_result = kurtosis_fractional_anisotropy(
            cti_params
        )
        assert np.allclose(kfa_result, kurtosis_fractional_anisotropy_result), (
            "The results of the kfa function from CorrelationTensorFit and the"
            "kurtosis_fractional_anisotropy function from dki.py are not equal"
        )
        # checking mean Kurtosis tensor
        mkt_result = ctiF.mkt(min_kurtosis=-3.0 / 7, max_kurtosis=10)
        mean_kurtosis_result = mean_kurtosis_tensor(
            cti_params, min_kurtosis=-3.0 / 7, max_kurtosis=10
        )
        assert np.allclose(mkt_result, mean_kurtosis_result), (
            "The results of the mkt function from CorrelationTensorFit and the"
            "mean_kurtosis_tensor function from dki.py are not equal."
        )
        # checking anisotropic source of kurtosis.
        K_aniso = ctiF.K_aniso
        variance_of_eigenvalues = []
        for tensor in DTD:
            evals_tensor, _ = decompose_tensor(tensor)
            variance_of_eigenvalues.append(np.var(evals_tensor))
        mean_variance_of_eigenvalues = np.mean(variance_of_eigenvalues)
        mean_D = np.trace(np.mean(DTD, axis=0)) / 3

        ground_truth_K_aniso = (6 / 5) * (mean_variance_of_eigenvalues / (mean_D**2))
        error_msg = (
            f"Calculated K_iso {K_aniso} for isotropicDTD does not match the "
            f"ground truth {ground_truth_K_aniso}"
        )
        assert np.isclose(K_aniso, ground_truth_K_aniso), error_msg

        # checking isotropic source of kurtosis.
        K_iso = ctiF.K_iso

        mean_diffusivities = []

        for tensor in DTD:
            evals_tensor, _ = decompose_tensor(tensor)
            mean_diffusivities.append(np.mean(evals_tensor))
        variance_of_mean_diffusivities = np.var(mean_diffusivities)
        mean_D = np.mean(mean_diffusivities)
        ground_truth_K_iso = 3 * variance_of_mean_diffusivities / (mean_D**2)

        error_msg = (
            f"Calculated K_iso {K_iso} for anisotropicDTD does not match the "
            f"ground truth {ground_truth_K_iso}"
        )
        assert np.isclose(K_iso, ground_truth_K_iso), error_msg

        # checking microscopic source of kurtosis
        ground_truth_K_micro = 0
        K_micro = ctiF.K_micro
        assert np.allclose(
            K_micro, ground_truth_K_micro
        ), "K_micro values don't match ground truth values"


def test_cti_errors():
    # first error of CTI module is if a unknown fit method is given
    assert_raises(ValueError, cti.CorrelationTensorModel, gtab1, gtab2, fit_method="")

    # second error of CTI module is if a min_signal is defined as negative
    assert_raises(ValueError, cti.CorrelationTensorModel, gtab1, gtab2, min_signal=-1)


def test_cti_design_matrix():
    A1 = design_matrix(gtab1, gtab2)
    A2 = design_matrix(gtab2, gtab1)
    # Check if the two matrices are the same
    assert np.allclose(A1, A2), (
        "The design matrices are not symmetric for different gradient"
        "directions order."
    )



from dipy.viz.horizon.tab.base import (
    HorizonTab,
    TabManager,
    build_checkbox,
    build_label,
    build_radio_button,
    build_slider,
    build_switcher,
)
from dipy.viz.horizon.tab.cluster import ClustersTab
from dipy.viz.horizon.tab.peak import PeaksTab
from dipy.viz.horizon.tab.roi import ROIsTab
from dipy.viz.horizon.tab.slice import SlicesTab
from dipy.viz.horizon.tab.surface import SurfaceTab

__all__ = [
    "HorizonTab",
    "TabManager",
    "ClustersTab",
    "PeaksTab",
    "ROIsTab",
    "SlicesTab",
    "build_label",
    "build_slider",
    "build_checkbox",
    "build_switcher",
    "SurfaceTab",
    "build_radio_button",
]



import importlib
import os
import sys
import warnings

import numpy as np
from numpy.testing import assert_almost_equal
import pytest

from dipy.data import get_fnames
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
original_backend = os.environ.get("DIPY_NN_BACKEND")
synb0_mod = None


def setup_module(module):
    """Set up environment variable for all tests in this module."""
    global synb0_mod
    os.environ["DIPY_NN_BACKEND"] = "tensorflow"
    with warnings.catch_warnings():
        msg = ".*uses TensorFlow.*install PyTorch.*"
        warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
        dipy_nn = importlib.reload(sys.modules["dipy.nn"])
    synb0_mod = dipy_nn.synb0


def teardown_module(module):
    """Restore the original environment variable after all tests in this module."""
    global synb0_mod
    if original_backend is not None:
        os.environ["DIPY_NN_BACKEND"] = original_backend
    else:
        del os.environ["DIPY_NN_BACKEND"]
    synb0_mod = None


@pytest.mark.skipif(not have_tf, reason="Requires TensorFlow")
def test_default_weights():
    file_names = get_fnames(name="synb0_test_data")
    input_arr1 = np.load(file_names[0])["b0"][0]
    input_arr2 = np.load(file_names[0])["T1"][0]
    target_arr = np.load(file_names[1])["arr_0"][0]

    synb0_model = synb0_mod.Synb0()
    synb0_model.fetch_default_weights(0)
    results_arr = synb0_model.predict(input_arr1, input_arr2, average=False)
    assert_almost_equal(results_arr, target_arr, decimal=1)


@pytest.mark.skipif(not have_tf, reason="Requires TensorFlow")
def test_default_weights_batch():
    file_names = get_fnames(name="synb0_test_data")
    input_arr1 = np.load(file_names[0])["b0"]
    input_arr2 = np.load(file_names[0])["T1"]
    target_arr = np.load(file_names[1])["arr_0"]
    synb0_model = synb0_mod.Synb0()
    synb0_model.fetch_default_weights(0)
    results_arr = synb0_model.predict(
        input_arr1, input_arr2, batch_size=2, average=False
    )
    assert_almost_equal(results_arr, target_arr, decimal=1)



#!python
import logging
import os
import sys

from dipy.utils.optpkg import optional_package
from dipy.workflows.flow_runner import run_flow

cli_flows = {
    "dipy_align_affine": ("dipy.workflows.align", "ImageRegistrationFlow"),
    "dipy_align_syn": ("dipy.workflows.align", "SynRegistrationFlow"),
    "dipy_apply_transform": ("dipy.workflows.align", "ApplyTransformFlow"),
    "dipy_buan_lmm": ("dipy.workflows.stats", "LinearMixedModelsFlow"),
    "dipy_buan_shapes": ("dipy.workflows.stats", "BundleShapeAnalysis"),
    "dipy_buan_profiles": ("dipy.workflows.stats", "BundleAnalysisTractometryFlow"),
    "dipy_bundlewarp": ("dipy.workflows.align", "BundleWarpFlow"),
    "dipy_classify_tissue": ("dipy.workflows.segment", "ClassifyTissueFlow"),
    "dipy_correct_motion": ("dipy.workflows.align", "MotionCorrectionFlow"),
    "dipy_correct_biasfield": ("dipy.workflows.nn", "BiasFieldCorrectionFlow"),
    "dipy_concatenate_tractograms": ("dipy.workflows.io", "ConcatenateTractogramFlow"),
    "dipy_convert_tractogram": ("dipy.workflows.io", "ConvertTractogramFlow"),
    "dipy_convert_tensors": ("dipy.workflows.io", "ConvertTensorsFlow"),
    "dipy_convert_sh": ("dipy.workflows.io", "ConvertSHFlow"),
    "dipy_denoise_nlmeans": ("dipy.workflows.denoise", "NLMeansFlow"),
    "dipy_denoise_lpca": ("dipy.workflows.denoise", "LPCAFlow"),
    "dipy_denoise_mppca": ("dipy.workflows.denoise", "MPPCAFlow"),
    "dipy_denoise_patch2self": ("dipy.workflows.denoise", "Patch2SelfFlow"),
    "dipy_evac_plus": ("dipy.workflows.nn", "EVACPlusFlow"),
    "dipy_fetch": ("dipy.workflows.io", "FetchFlow"),
    "dipy_fit_csa": ("dipy.workflows.reconst", "ReconstQBallBaseFlow"),
    "dipy_fit_csd": ("dipy.workflows.reconst", "ReconstCSDFlow"),
    "dipy_fit_dki": ("dipy.workflows.reconst", "ReconstDkiFlow"),
    "dipy_fit_dti": ("dipy.workflows.reconst", "ReconstDtiFlow"),
    "dipy_fit_dsi": ("dipy.workflows.reconst", "ReconstDsiFlow"),
    "dipy_fit_dsid": ("dipy.workflows.reconst", "ReconstDsiFlow"),
    "dipy_fit_forecast": ("dipy.workflows.reconst", "ReconstForecastFlow"),
    "dipy_fit_gqi": ("dipy.workflows.reconst", "ReconstGQIFlow"),
    "dipy_fit_ivim": ("dipy.workflows.reconst", "ReconstIvimFlow"),
    "dipy_fit_mapmri": ("dipy.workflows.reconst", "ReconstMAPMRIFlow"),
    "dipy_fit_opdt": ("dipy.workflows.reconst", "ReconstQBallBaseFlow"),
    "dipy_fit_qball": ("dipy.workflows.reconst", "ReconstQBallBaseFlow"),
    "dipy_fit_sdt": ("dipy.workflows.reconst", "ReconstSDTFlow"),
    "dipy_fit_sfm": ("dipy.workflows.reconst", "ReconstSFMFlow"),
    "dipy_gibbs_ringing": ("dipy.workflows.denoise", "GibbsRingingFlow"),
    "dipy_horizon": ("dipy.workflows.viz", "HorizonFlow"),
    "dipy_info": ("dipy.workflows.io", "IoInfoFlow"),
    "dipy_extract_b0": ("dipy.workflows.io", "ExtractB0Flow"),
    "dipy_extract_shell": ("dipy.workflows.io", "ExtractShellFlow"),
    "dipy_extract_volume": ("dipy.workflows.io", "ExtractVolumeFlow"),
    "dipy_labelsbundles": ("dipy.workflows.segment", "LabelsBundlesFlow"),
    "dipy_math": ("dipy.workflows.io", "MathFlow"),
    "dipy_mask": ("dipy.workflows.mask", "MaskFlow"),
    "dipy_median_otsu": ("dipy.workflows.segment", "MedianOtsuFlow"),
    "dipy_nifti2pam": ("dipy.workflows.io", "NiftisToPamFlow"),
    "dipy_pam2nifti": ("dipy.workflows.io", "PamToNiftisFlow"),
    "dipy_recobundles": ("dipy.workflows.segment", "RecoBundlesFlow"),
    "dipy_reslice": ("dipy.workflows.align", "ResliceFlow"),
    "dipy_sh_convert_mrtrix": ("dipy.workflows.io", "ConvertSHFlow"),
    "dipy_slr": ("dipy.workflows.align", "SlrWithQbxFlow"),
    "dipy_snr_in_cc": ("dipy.workflows.stats", "SNRinCCFlow"),
    "dipy_split": ("dipy.workflows.io", "SplitFlow"),
    "dipy_tensor2pam": ("dipy.workflows.io", "TensorToPamFlow"),
    "dipy_track": ("dipy.workflows.tracking", "LocalFiberTrackingPAMFlow"),
    "dipy_track_pft": ("dipy.workflows.tracking", "PFTrackingPAMFlow"),
}


def run():
    """Run scripts located in pyproject.toml."""
    script_name = os.path.basename(sys.argv[0])
    mod_name, flow_name = cli_flows.get(script_name, (None, None))
    if mod_name is None:
        print(f"Flow: {script_name} not Found in DIPY")
        print(f"Available flows: {', '.join(cli_flows.keys())}")
        sys.exit(1)
    mod, _, _ = optional_package(mod_name)

    if script_name in ["dipy_sh_convert_mrtrix"]:
        logging.warning(
            "`dipy_sh_convert_mrtrix` CLI is deprecated since DIPY 1.11.0. It will be "
            "removed on later release. Please use the `dipy_convert_sh` CLI instead.",
        )

    extra_args = {}
    if script_name == "dipy_fit_dsid":
        extra_args = {
            "remove_convolution": {
                "dest": "remove_convolution",
                "action": "store_true",
                "default": True,
            }
        }
    elif script_name == "dipy_fit_csa":
        extra_args = {
            "method": {
                "action": "store",
                "dest": "method",
                "metavar": "string",
                "default": "csa",
            }
        }
    elif script_name == "dipy_fit_opdt":
        extra_args = {
            "method": {
                "action": "store",
                "dest": "method",
                "metavar": "string",
                "default": "opdt",
            }
        }
    elif script_name == "dipy_fit_qball":
        extra_args = {
            "method": {
                "action": "store",
                "dest": "method",
                "metavar": "string",
                "default": "qball",
            }
        }
    run_flow(getattr(mod, flow_name)(), extra_args=extra_args)



"""
===========================================================================
Reconstruction of the diffusion signal with the kurtosis tensor model (DKI)
===========================================================================

Diffusional Kurtosis Imaging (DKI) is an expansion of the Diffusion Tensor
Imaging (DTI) model
(see :ref:`sphx_glr_examples_built_reconstruction_reconst_dti.py`). In
addition to the Diffusion Tensor (DT), DKI quantifies the degree to which water
diffusion in biological tissues is non-Gaussian using the Kurtosis Tensor (KT)
:footcite:p:`Jensen2005`.

Measurements of non-Gaussian diffusion from DKI are of interest because they
were shown to provide extra information about microstructural alterations in
both health and disease (for a review see our
paper :footcite:p:`NetoHenriques2021a`). Moreover, in contrast to DTI, DKI can
provide metrics of tissue microscopic heterogeneity that are less sensitive to
confounding effects in the orientation of tissue components, thus providing
better characterization in general white matter configurations (including
regions of fibers crossing, fanning, and/or dispersing) and gray matter
:footcite:p:`NetoHenriques2015`, :footcite:p:`NetoHenriques2021a`. Although DKI
aims primarily to quantify the degree of non-Gaussian diffusion without
establishing concrete biophysical assumptions, DKI can also be related to
microstructural models to infer specific biophysical parameters (e.g., the
density of axonal fibers) - this aspect will be more closely explored in
:ref:sphx_glr_examples_built_reconstruction_reconst_dki_micro.py. For
additional information on DKI and its practical implementation within DIPY,
refer to :footcite:p:`NetoHenriques2021a`.

Below, we introduce a concise theoretical background of DKI and demonstrate
its fitting process using DIPY. We'll also guide you through the fitting
process of DKI using DIPY, demonstrating how to effectively apply this
technique. Furthermore, we discuss the various diffusion metrics that can be
derived from DKI, providing insight into their practical significance and
applications. Additionally, we address strategies to mitigate common artifacts,
such as implausible negative kurtosis estimates, which manifest as 'black'
voxels or holes in DKI maps. These artifacts can compromise the accuracy of
the DKI analysis, and we'll offer solutions to ensure more reliable results.

Theory
======

The DKI model expresses the diffusion-weighted signal as:

.. math::

    S(n,b)=S_{0}e^{-bD(n)+\\frac{1}{6}b^{2}D(n)^{2}K(n)}

where $\\mathbf{b}$ is the applied diffusion weighting (which is dependent on
the measurement parameters), $S_0$ is the signal in the absence of diffusion
gradient sensitization, $\\mathbf{D(n)}$ is the value of diffusion along
direction $\\mathbf{n}$, and $\\mathbf{K(n)}$ is the value of kurtosis along
direction $\\mathbf{n}$. The directional diffusion $\\mathbf{D(n)}$ and kurtosis
$\\mathbf{K(n)}$ can be related to the diffusion tensor (DT) and kurtosis tensor
(KT) using the following equations:

.. math::
     D(n)=\\sum_{i=1}^{3}\\sum_{j=1}^{3}n_{i}n_{j}D_{ij}

and

.. math::
     K(n)=\\frac{MD^{2}}{D(n)^{2}}\\sum_{i=1}^{3}\\sum_{j=1}^{3}\\sum_{k=1}^{3}
     \\sum_{l=1}^{3}n_{i}n_{j}n_{k}n_{l}W_{ijkl}

where $D_{ij}$ are the elements of the second-order DT, and $W_{ijkl}$ the
elements of the fourth-order KT and $MD$ is the mean diffusivity. As the DT,
KT has antipodal symmetry and thus only 15 Wijkl elements are needed to fully
characterize the KT:

.. math::
   \\begin{matrix} ( & W_{xxxx} & W_{yyyy} & W_{zzzz} & W_{xxxy} & W_{xxxz}
                    & ... \\\\
                    & W_{xyyy} & W_{yyyz} & W_{xzzz} & W_{yzzz} & W_{xxyy}
                    & ... \\\\
                    & W_{xxzz} & W_{yyzz} & W_{xxyz} & W_{xyyz} & W_{xyzz}
                    & & )\\end{matrix}

DKI fitting in DIPY
===================

In the following example we show how to fit the diffusion kurtosis model on
diffusion-weighted multi-shell datasets and how to estimate diffusion kurtosis
based statistics.

First, we import all relevant modules:
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.denoise.localpca import mppca
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
import dipy.reconst.dki as dki
import dipy.reconst.dti as dti
from dipy.segment.mask import median_otsu
from dipy.viz.plotting import compare_maps

###############################################################################
# DKI requires multi-shell data, i.e. data acquired from more than one
# non-zero b-value. Here, we use fetch to download a multi-shell dataset which
# was kindly provided by Hansen and Jespersen (more details about the data are
# provided in their paper :footcite:p:`Hansen2016a`). The total size of the
# downloaded data is 192 MBytes, however you only need to fetch it once.

fraw, fbval, fbvec, t1_fname = get_fnames(name="cfin_multib")

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# Function ``get_fnames`` downloads and outputs the paths of the data,
# ``load_nifti`` returns the data as a nibabel Nifti1Image object, and
# ``read_bvals_bvecs`` loads the arrays containing the information about the
# b-values and b-vectors. These later arrays are converted to the
# GradientTable object required for Dipy_'s data reconstruction.
#
# The downloaded dataset was acquired with an unusually large number of
# b-values. To run this example with acquisitions that are more common in
# practice, we select below data for three non-zero b-values (if you want to
# run this example with the full data extent, skip the following lines of
# code)

bval_sel = np.zeros_like(gtab.bvals)
bval_sel[bvals == 0] = 1
bval_sel[bvals == 600] = 1
bval_sel[bvals == 1000] = 1
bval_sel[bvals == 2000] = 1

data = data[..., bval_sel == 1]
gtab = gradient_table(bvals[bval_sel == 1], bvecs=bvecs[bval_sel == 1])

###############################################################################
# Before fitting the data, we perform some data pre-processing. We first
# compute a brain mask to avoid unnecessary calculations on the background
# of the image.

datamask, mask = median_otsu(
    data, vol_idx=[0, 1], median_radius=4, numpass=2, autocrop=False, dilate=1
)

###############################################################################
# Since the diffusion kurtosis model involves the estimation of a large number
# of parameters :footcite:p:`Tax2015` and since the non-Gaussian components of
# the diffusion signal are more sensitive to artifacts
# :footcite:p:`NetoHenriques2012`, :footcite:p:`Tabesh2011`, it might be
# favorable to suppress the effects of noise and artifacts before diffusion
# kurtosis fitting. In this example, the effects of noise are suppressed using
# the Marcenko-Pastur (MP)-PCA algorithm (for more information, see
# :ref:sphx_glr_examples_built_preprocessing_denoise_mppca.py). Processing
# MP-PCA may take a while - for illustration purposes, you can skip this step.
# However, note that if you don't denoise your data, DKI reconstructions may
# be corrupted by a large percentage of implausible DKI estimates (see below
# for more information on this issue).

data = mppca(data, patch_radius=[3, 3, 3])

###############################################################################
# Now that we have loaded and pre-processed the data we can go forward
# with DKI fitting. For this, the DKI model is first defined for the data's
# GradientTable object by instantiating the DiffusionKurtosisModel object in
# the following way:

dkimodel = dki.DiffusionKurtosisModel(gtab)

###############################################################################
# To fit the data using the defined model object, we call the ``fit`` function
# of this object. For the purpose of this example, we will only fit a
# single slice of the data:

dkifit = dkimodel.fit(data[:, :, 9:10], mask=mask[:, :, 9:10])

###############################################################################
# The fit method creates a DiffusionKurtosisFit object, which contains all the
# diffusion and kurtosis fitting parameters and other DKI attributes. For
# instance, since the diffusion kurtosis model estimates the diffusion tensor,
# all standard diffusion tensor statistics can be computed from the
# DiffusionKurtosisFit instance. For example, we can extract the fractional
# anisotropy (FA), the mean diffusivity (MD), the radial diffusivity (RD) and
# the axial diffusivity (AD) from the DiffusionKurtosisiFit instance. Of
# course, these measures can also be computed from DIPY's ``TensorModel`` fit,
# and should be analogous; however, theoretically, the diffusion statistics
# from the kurtosis model are expected to have better accuracy, since DKI's
# diffusion tensor are decoupled from higher order terms effects
# :footcite:p:`Veraart2011`, :footcite:p:`NetoHenriques2021a`. Below we compare
# the FA, MD, AD, and RD, computed from both DTI and DKI.

tenmodel = dti.TensorModel(gtab)
tenfit = tenmodel.fit(data[:, :, 9:10], mask=mask[:, :, 9:10])

fits = [tenfit, dkifit]
maps = ["fa", "md", "rd", "ad"]
fit_labels = ["DTI", "DKI"]
map_kwargs = [{"vmax": 0.7}, {"vmax": 2e-3}, {"vmax": 2e-3}, {"vmax": 2e-3}]
compare_maps(
    fits,
    maps,
    fit_labels=fit_labels,
    map_kwargs=map_kwargs,
    filename="Diffusion_tensor_measures_from_DTI_and_DKI.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Diffusion tensor measures obtained from the diffusion tensor estimated
# from DKI (upper panels) and DTI (lower panels).
#
#
# DTI's diffusion estimates present lower values than DKI's estimates,
# showing that DTI's diffusion measurements are underestimated by higher
# order effects (for detailed discussion on this see
# :footcite:p:`NetoHenriques2021a`.
#
# In addition to the standard diffusion statistics, the DiffusionKurtosisFit
# instance can be used to estimate the non-Gaussian measures of mean kurtosis
# (MK), the radial kurtosis (RK) and the axial kurtosis (AK).

maps = ["mk", "rk", "ak"]
compare_maps(
    [dkifit],
    maps,
    fit_labels=["DKI"],
    map_kwargs={"vmin": 0, "vmax": 1.5},
    filename="DKI_standard_measures.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# DKI standard kurtosis measures.
#
#
# The non-Gaussian behaviour of the diffusion signal is expected to be higher
# when tissue water is confined by multiple compartments. MK is, therefore,
# higher in white matter since it is highly compartmentalized by myelin
# sheaths. These water diffusion compartmentalization is expected to be more
# pronounced perpendicularly to white matter fibers and thus the RK map
# presents higher amplitudes than the AK map.
#
# Mitigating 'Black' Voxels / Holes in DKI metrics
# ================================================
#
# It is important to note that kurtosis estimates might present implausible
# negative estimates in deep white matter regions that will manifest as
# 'Black' voxels or holes in DKI metrics (e.g. see the band of dark voxels in
# the  RK map above). These negative kurtosis values are artifactual and might
# be induced by:
# 1) low radial diffusivities of aligned white matter - since it is very hard
# to capture non-Gaussian information in radial direction due to its low
# diffusion decays, radial kurtosis estimates (and consequently the mean
# kurtosis estimates) might have low robustness and tendency to exhibit
# negative values :footcite:p:`NetoHenriques2012`, :footcite:p:`Tabesh2011`;
# 2) Gibbs artifacts - MRI images might be corrupted by signal oscillation
# artifact between tissue's edges if an inadequate number of high frequencies
# of the k-space is sampled. These oscillations might have different signs on
# images acquired with different diffusion-weighted and inducing negative
# biases in kurtosis parametric maps :footcite:p:`Perrone2015`,
# :footcite:p:`NetoHenriques2018`. 3) Underestimation of b0 signals - Due to
# physiological or noise artifacts, the signals acquired at b-value=0 may be
# artifactually lower than the diffusion-weighted signals acquired for the
# different b-values. In this case, the log diffusion-weighted signal decay may
# appear to be concave rather than showing to be convex (as one would typically
# expect), leading to negative kurtosis value estimates.
#
# Given the above, one can try to suppress the 'Black' voxel / holes in DKI
# metrics by:
# 1) using more advanced noise and artifact suppression algorithms, e.g.,
# as mentioned above, the MP-PCA denoising
# (:ref:`sphx_glr_examples_built_preprocessing_denoise_mppca.py`), other
# denoising alternatives such as Patch2self
# (:ref:`sphx_glr_examples_built_preprocessing_denoise_patch2self.py`) or
# incorporating methods for Gibbs Artifact Unringing
# (:ref:`sphx_glr_examples_built_preprocessing_denoise_gibbs.py`)
# algorithms.
# 2) computing the kurtosis values from powder-averaged diffusion-weighted
# signals which are known to be less sensitive to implausible negative
# estimates. The details on how to compute the kurtosis from powder-averaged
# signals in DIPY are described in the following tutorial
# (:ref:`sphx_glr_examples_built_reconstruction_reconst_msdki.py`).
# 3) computing alternative definitions of mean and radial kurtosis such as
# the mean kurtosis tensor (MKT) and radial tensor kurtosis (RTK) metrics (see
# below).
# 4) constrained optimization to ensure that the fitted parameters
# are physically plausible :footcite:p:`DelaHaije2020` (see below).
#
# Alternative DKI metrics
# =======================
#
# In addition to the standard mean, axial, and radial kurtosis metrics shown
# above, alternative metrics can be computed from DKI, e.g.:
# 1) the mean kurtosis tensor (MKT) - defined as the trace of the kurtosis
# tensor - is a quantity that provides a contrast similar to the standard MK
# but it is more robust to noise artifacts :footcite:p:`Hansen2013`,
# :footcite:p:`NetoHenriques2021a`. 2) the radial tensor kurtosis (RTK) provides
# an alternative definition to standard radial kurtosis (RK) that, as MKT, is
# more robust to noise artifacts :footcite:p:`Hansen2013`.
# 3) the kurtosis fractional anisotropy (KFA) that quantifies the anisotropy of
# the kurtosis tensor :footcite:p:`Glenn2015`, which provides different
# information than the FA measures from the diffusion tensor.
#
# These measures are computed and illustrated below:

compare_maps(
    [dkifit],
    ["mkt", "rtk", "kfa"],
    fit_labels=["DKI"],
    map_kwargs=[
        {"vmin": 0, "vmax": 1.5},
        {"vmin": 0, "vmax": 1.5},
        {"vmin": 0, "vmax": 1},
    ],
    filename="Alternative_DKI_metrics.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Alternative DKI measures.
#
#
# Constrained optimization for DKI
# ================================
#
# When instantiating the DiffusionKurtosisModel, the model can be set up to use
# constraints with the option `fit_method='CLS'` (for ordinary least squares)
# or with `fit_method='CWLS'` (for weighted least squares). Constrained fitting
# takes more time than unconstrained fitting, but is generally recommended to
# prevent physically implausible parameter estimates
# :footcite:p:`DelaHaije2020`. For performance purposes it is recommended to use
# the MOSEK solver (https://www.mosek.com/) by setting ``cvxpy_solver='MOSEK'``.
# Different solvers can differ greatly in terms of runtime and solution
# accuracy, and in some cases solvers may show warnings about convergence or
# recommended option settings.
#
# .. note::
#    In certain atypical scenarios, the DKI+ constraints could potentially be
#    too restrictive. Always check the results of a constrained fit with their
#    unconstrained counterpart to verify that there are no unexpected
#    qualitative differences.

dkimodel_plus = dki.DiffusionKurtosisModel(gtab, fit_method="CLS")
dkifit_plus = dkimodel_plus.fit(data[:, :, 9:10], mask=mask[:, :, 9:10])

###############################################################################
# We can now compare the kurtosis measures obtained with the constrained fit to
# the measures obtained before, where we see that many of the artifactual
# voxels have now been corrected. In particular outliers caused by pure noise
# -- instead of for example acquisition artifacts -- can be corrected with
# this method.

compare_maps(
    [dkifit, dkifit_plus],
    ["mkt", "rtk", "ak"],
    fit_labels=["DKI", "DKI+"],
    map_kwargs={"vmin": 0, "vmax": 1.5},
    filename="Alternative_DKI_measures_comparison_to_DKIplus.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# DKI standard kurtosis measures obtained with constrained optimization.
#
# References
# ----------
#
# .. footbibliography::
#



from pathlib import Path

from dipy.viz.horizon.tab import HorizonTab, build_checkbox, build_label, build_slider


class SurfaceTab(HorizonTab):
    def __init__(self, visualizer, tab_name, file_name):
        """Surface Tab.

        Parameters
        ----------
        visualizer : SurfaceVisualizer
        id : int
        """
        super().__init__()
        self._visualizer = visualizer
        self._tab_id = 0

        self._name = tab_name
        self._file_name = Path(file_name or tab_name).name

        self._actor_toggle = build_checkbox(
            labels=[""], checked_labels=[""], on_change=self._toggle_actors
        )

        self._surface_opacity_label, self._surface_opacity = build_slider(
            initial_value=1.0,
            max_value=1.0,
            text_template="{ratio:.0%}",
            on_change=self._change_opacity,
            label="Opacity",
        )

        self._file_label = build_label(
            text="Filename",
        )
        self._file_name_label = build_label(text=self._file_name)

        self._register_elements(
            self._actor_toggle,
            self._surface_opacity_label,
            self._surface_opacity,
            self._file_label,
            self._file_name_label,
        )

    def _change_opacity(self, slider):
        """Change opacity value according to slider changed.

        Parameters
        ----------
        slider : LineSlider2D
        """
        self._surface_opacity.selected_value = slider.value
        self._update_opacities()

    def _update_opacities(self):
        """Update opacities of visible actors based on selected values."""
        for actor in self.actors:
            actor.GetProperty().SetOpacity(self._surface_opacity.selected_value)

    def build(self, tab_id):
        """Build all the elements under the tab.

        Parameters
        ----------
        tab_id : int
            Id of the tab.
        """
        self._tab_id = tab_id

        y_pos = 0.85
        self._actor_toggle.position = (0.02, y_pos)
        self._surface_opacity_label.position = (0.05, y_pos)
        self._surface_opacity.position = (0.10, y_pos)

        y_pos = 0.65
        self._file_label.position = (0.05, y_pos)
        self._file_name_label.position = (0.13, y_pos)

    @property
    def name(self):
        """Name of the tab.

        Returns
        -------
        str
        """
        return self._name

    @property
    def actors(self):
        """Actors controlled by this tab."""
        return self._visualizer.actors



# cython: boundscheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: overflowcheck=False


import numpy as np
cimport numpy as cnp

import cython
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy
from libc.math cimport cos, M_PI, INFINITY

from dipy.core.math cimport f_max, f_array_min
from dipy.reconst.recspeed cimport local_maxima_c, search_descending_c, remove_similar_vertices_c


cdef cnp.uint16_t peak_directions_c(
    double[:] odf,
    double[:, ::1] sphere_vertices,
    cnp.uint16_t[:, ::1] sphere_edges,
    double relative_peak_threshold,
    double min_separation_angle,
    bint is_symmetric,
    double[:, ::1] out_directions,
    double[::1] out_values,
    cnp.npy_intp[::1] out_indices,
    double[:, :] unique_vertices,
    cnp.uint16_t[:] mapping,
    cnp.uint16_t[:] index
) noexcept nogil:
    """Get the directions of odf peaks.

    Peaks are defined as points on the odf that are greater than at least one
    neighbor and greater than or equal to all neighbors. Peaks are sorted in
    descending order by their values then filtered based on their relative size
    and spacing on the sphere. An odf may have 0 peaks, for example if the odf
    is perfectly isotropic.

    Parameters
    ----------
    odf : 1d ndarray
        The odf function evaluated on the vertices of `sphere`
    sphere_vertices : (N,3) ndarray
        The sphere vertices providing discrete directions for evaluation.
    sphere_edges : (N, 2) ndarray

    relative_peak_threshold : float in [0., 1.]
        Only peaks greater than ``min + relative_peak_threshold * scale`` are
        kept, where ``min = max(0, odf.min())`` and
        ``scale = odf.max() - min``.
    min_separation_angle : float in [0, 90]
        The minimum distance between directions. If two peaks are too close
        only the larger of the two is returned.
    is_symmetric : bool, optional
        If True, v is considered equal to -v.
    out_directions : (N, 3) ndarray
        The directions of the peaks, N vertices for sphere, one for each peak
    out_values : (N,) ndarray
        The peak values
    out_indices: (N,) ndarray
        The peak indices of the directions on the sphere
    unique_vertices : (N, 3) ndarray
        The unique vertices
    mapping  : (N,) ndarray
        For each element ``vertices[i]`` ($i \in 0..N-1$), the index $j$ to a
        vertex in `unique_vertices` that is less than `theta` degrees from
        ``vertices[i]``.
    index : (N,) ndarray
        `index` gives the reverse of `mapping`.  For each element
        ``unique_vertices[j]`` ($j \in 0..M-1$), the index $i$ to a vertex in
        `vertices` that is less than `theta` degrees from
        ``unique_vertices[j]``.  If there is more than one element of
        `vertices` that is less than theta degrees from `unique_vertices[j]`,
        return the first (lowest index) matching value.
    Returns
    -------
    n_unique : int
        The number of unique peaks

    Notes
    -----
    If the odf has any negative values, they will be clipped to zeros.

    """
    cdef cnp.npy_intp i, n, idx
    cdef long count
    cdef double odf_min
    cdef cnp.npy_intp num_vertices = sphere_vertices.shape[0]
    cdef double* tmp_buffer
    cdef cnp.uint16_t n_unique = 0

    count = local_maxima_c(odf, sphere_edges, out_values, out_indices)

    # If there is only one peak return
    if count == 0:
        return 0
    elif count == 1:
        return 1

    odf_min = f_array_min[double](&odf[0], num_vertices)
    odf_min = f_max[double](odf_min, 0.0)

    # because of the relative threshold this algorithm will give the same peaks
    # as if we divide (values - odf_min) with (odf_max - odf_min) or not so
    # here we skip the division to increase speed
    tmp_buffer = <double*>malloc(count * sizeof(double))
    memcpy(tmp_buffer, &out_values[0], count * sizeof(double))

    for i in range(count):
        tmp_buffer[i] -= odf_min

    # Remove small peaks
    n = search_descending_c[cython.double](tmp_buffer, <cnp.npy_intp>count, relative_peak_threshold)

    for i in range(n):
        idx = out_indices[i]
        out_directions[i, :] = sphere_vertices[idx, :]

    # Remove peaks too close together
    remove_similar_vertices_c(
        out_directions[:n, :],
        min_separation_angle,
        is_symmetric,
        0,  # return_mapping
        1,  # return_index
        unique_vertices[:n, :],
        mapping[:n],
        index[:n],
        &n_unique
    )

    # Update final results
    for i in range(n_unique):
        idx = index[i]
        out_directions[i, :] = unique_vertices[i, :]
        out_values[i] = out_values[idx]
        out_indices[i] = out_indices[idx]

    free(tmp_buffer)
    return n_unique


def peak_directions(
    double[:] odf,
    sphere,
    *,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    bint is_symmetric=True,
):
    """Get the directions of odf peaks.

    Peaks are defined as points on the odf that are greater than at least one
    neighbor and greater than or equal to all neighbors. Peaks are sorted in
    descending order by their values then filtered based on their relative size
    and spacing on the sphere. An odf may have 0 peaks, for example if the odf
    is perfectly isotropic.

    Parameters
    ----------
    odf : 1d ndarray
        The odf function evaluated on the vertices of `sphere`
    sphere : Sphere
        The Sphere providing discrete directions for evaluation.
    relative_peak_threshold : float in [0., 1.]
        Only peaks greater than ``min + relative_peak_threshold * scale`` are
        kept, where ``min = max(0, odf.min())`` and
        ``scale = odf.max() - min``.
    min_separation_angle : float in [0, 90]
        The minimum distance between directions. If two peaks are too close
        only the larger of the two is returned.
    is_symmetric : bool, optional
        If True, v is considered equal to -v.

    Returns
    -------
    directions : (N, 3) ndarray
        N vertices for sphere, one for each peak
    values : (N,) ndarray
        peak values
    indices : (N,) ndarray
        peak indices of the directions on the sphere

    Notes
    -----
    If the odf has any negative values, they will be clipped to zeros.

    """
    cdef double[:, ::1] vertices = sphere.vertices
    cdef cnp.uint16_t[:, ::1] edges = sphere.edges

    cdef cnp.npy_intp num_vertices = sphere.vertices.shape[0]
    cdef cnp.uint16_t n_unique = 0

    cdef double[:, ::1] directions_out = np.zeros((num_vertices, 3), dtype=np.float64)
    cdef double[::1] values_out = np.zeros(num_vertices, dtype=np.float64)
    cdef cnp.npy_intp[::1] indices_out = np.zeros(num_vertices, dtype=np.intp)
    cdef cnp.float64_t[:, ::1] unique_vertices = np.empty((num_vertices, 3), dtype=np.float64)
    cdef cnp.uint16_t[::1] mapping = None
    cdef cnp.uint16_t[::1] index = np.empty(num_vertices, dtype=np.uint16)

    n_unique = peak_directions_c(
        odf, vertices, edges, relative_peak_threshold, min_separation_angle, is_symmetric,
        directions_out, values_out, indices_out, unique_vertices, mapping, index
    )

    if n_unique == 0:
        return np.zeros((0, 3)), np.zeros(0), np.zeros(0, dtype=int)
    elif n_unique == 1:
        return (
            sphere.vertices[np.asarray(indices_out[:n_unique]).astype(int)],
            np.asarray(values_out[:n_unique]),
            np.asarray(indices_out[:n_unique]),
        )
    return (
        np.asarray(directions_out[:n_unique, :]),
        np.asarray(values_out[:n_unique]),
        np.asarray(indices_out[:n_unique]),
    )


import numpy as np

from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from fury.actor import surface as surface_actor


class SurfaceVisualizer:
    def __init__(self, surface, scene, color):
        self._vertices, self._faces = surface

        self._surface_actor = surface_actor(
            self._vertices,
            faces=self._faces,
            colors=np.full((self._vertices.shape[0], 3), color),
        )

        scene.add(self._surface_actor)

    @property
    def actors(self):
        return [self._surface_actor]



import warnings

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises
from scipy.ndimage import binary_dilation, generate_binary_structure, median_filter

from dipy.data import get_fnames
from dipy.io.image import load_nifti_data
from dipy.segment.mask import (
    applymask,
    bounding_box,
    crop,
    median_otsu,
    multi_median,
    otsu,
)


def test_mask():
    vol = np.zeros((30, 30, 30))
    vol[15, 15, 15] = 1
    struct = generate_binary_structure(3, 1)
    voln = binary_dilation(vol, structure=struct, iterations=4).astype("f4")
    initial = np.sum(voln > 0)
    mask = voln.copy()
    thresh = otsu(mask)
    mask = mask > thresh
    initial_otsu = np.sum(mask > 0)
    assert_equal(initial_otsu, initial)

    mins, maxs = bounding_box(mask)
    voln_crop = crop(mask, mins, maxs)
    initial_crop = np.sum(voln_crop > 0)
    assert_equal(initial_crop, initial)

    applymask(voln, mask)
    final = np.sum(voln > 0)
    assert_equal(final, initial)

    # Test multi_median.
    img = np.arange(25).reshape(5, 5)
    img_copy = img.copy()
    medianradius = 2
    median_test = multi_median(img, medianradius, 3)
    assert_equal(img, img_copy)

    medarr = np.ones_like(img.shape) * ((medianradius * 2) + 1)
    median_control = median_filter(img, medarr)
    median_control = median_filter(median_control, medarr)
    median_control = median_filter(median_control, medarr)
    assert_equal(median_test, median_control)


def test_bounding_box():
    vol = np.zeros((100, 100, 50), dtype=int)

    # Check the more usual case
    vol[10:90, 11:40, 5:33] = 3
    mins, maxs = bounding_box(vol)
    assert_equal(mins, [10, 11, 5])
    assert_equal(maxs, [90, 40, 33])

    # Check a 2d case
    mins, maxs = bounding_box(vol[10])
    assert_equal(mins, [11, 5])
    assert_equal(maxs, [40, 33])

    vol[:] = 0
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        # Trigger a warning.
        num_warns = len(w)
        mins, maxs = bounding_box(vol)
        # Assert number of warnings has gone up by 1
        assert_equal(len(w), num_warns + 1)

        # Check that an empty array returns zeros for both min & max
        assert_equal(mins, [0, 0, 0])
        assert_equal(maxs, [0, 0, 0])

        # Check the 2d case
        mins, maxs = bounding_box(vol[0])
        assert_equal(len(w), num_warns + 2)
        assert_equal(mins, [0, 0])
        assert_equal(maxs, [0, 0])


def test_median_otsu():
    fname = get_fnames(name="S0_10")
    data = load_nifti_data(fname)
    data = np.squeeze(data.astype("f8"))
    dummy_mask = data > data.mean()
    data_masked, mask = median_otsu(
        data, median_radius=3, numpass=2, autocrop=False, vol_idx=None, dilate=None
    )
    assert_equal(mask.sum() < dummy_mask.sum(), True)
    data2 = np.zeros(data.shape + (2,))
    data2[..., 0] = data
    data2[..., 1] = data

    data2_masked, mask2 = median_otsu(
        data2, median_radius=3, numpass=2, autocrop=False, vol_idx=[0, 1], dilate=None
    )
    assert_almost_equal(mask.sum(), mask2.sum())

    _, mask3 = median_otsu(
        data2, median_radius=3, numpass=2, autocrop=False, vol_idx=[0, 1], dilate=1
    )
    assert_equal(mask2.sum() < mask3.sum(), True)

    _, mask4 = median_otsu(
        data2, median_radius=3, numpass=2, autocrop=False, vol_idx=[0, 1], dilate=2
    )
    assert_equal(mask3.sum() < mask4.sum(), True)

    # For 4D volumes, can't call without vol_idx input:
    assert_raises(ValueError, median_otsu, data2)
